# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=missing-function-docstring

"""Test the StarPreRouting pass"""

import unittest

from test import QiskitTestCase
import ddt

from qiskit.circuit.library import QFT
from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.converters import circuit_to_dag, dag_to_circuit
from qiskit.quantum_info import Operator
from qiskit.transpiler.passes import VF2Layout, ApplyLayout, SabreSwap, SabreLayout
from qiskit.transpiler.passes.layout.vf2_utils import build_interaction_graph
from qiskit.transpiler.passes.routing.star_prerouting import StarPreRouting
from qiskit.transpiler.coupling import CouplingMap
from qiskit.transpiler.passmanager import PassManager
from qiskit.transpiler.preset_passmanagers import generate_preset_pass_manager
from qiskit.utils.optionals import HAS_AER


@ddt.ddt
class TestStarPreRouting(QiskitTestCase):
    """Tests the StarPreRouting pass"""

    def test_simple_ghz_dagcircuit(self):
        qc = QuantumCircuit(5)
        qc.h(0)
        qc.cx(0, range(1, 5))
        dag = circuit_to_dag(qc)
        new_dag = StarPreRouting().run(dag)
        new_qc = dag_to_circuit(new_dag)

        expected = QuantumCircuit(5)
        expected.h(0)
        expected.cx(0, 1)
        expected.cx(0, 2)
        expected.swap(0, 2)
        expected.cx(2, 3)
        expected.swap(2, 3)
        expected.cx(3, 4)
        # expected.swap(3,4)

        self.assertTrue(Operator(expected).equiv(Operator(new_qc)))

    def test_simple_ghz_dagdependency(self):
        qc = QuantumCircuit(5)
        qc.h(0)
        qc.cx(0, range(1, 5))

        pm = generate_preset_pass_manager(optimization_level=3, seed_transpiler=42)
        pm.init += StarPreRouting()

        result = pm.run(qc)

        self.assertTrue(Operator.from_circuit(result).equiv(Operator(qc)))

    def test_double_ghz_dagcircuit(self):
        qc = QuantumCircuit(10)
        qc.h(0)
        qc.cx(0, range(1, 5))
        qc.h(9)
        qc.cx(9, range(8, 4, -1))

        pm = generate_preset_pass_manager(optimization_level=3, seed_transpiler=42)
        pm.init += StarPreRouting()
        new_qc = pm.run(qc)

        self.assertTrue(Operator.from_circuit(new_qc).equiv(Operator(qc)))

    def test_double_ghz_dagdependency(self):
        qc = QuantumCircuit(10)
        qc.h(0)
        qc.cx(0, range(1, 5))
        qc.h(9)
        qc.cx(9, range(8, 4, -1))
        pm = generate_preset_pass_manager(optimization_level=3, seed_transpiler=42)
        pm.init += StarPreRouting()
        new_qc = pm.run(qc)

        self.assertTrue(Operator(qc).equiv(Operator.from_circuit(new_qc)))

    def test_mixed_double_ghz_dagdependency(self):
        """Shows off the power of using commutation analysis."""
        qc = QuantumCircuit(4)
        qc.cx(0, 1)
        qc.cx(0, 2)

        qc.cx(3, 1)
        qc.cx(3, 2)

        qc.cx(0, 1)
        qc.cx(0, 2)

        qc.cx(3, 1)
        qc.cx(3, 2)

        qc.cx(0, 1)
        qc.cx(0, 2)

        qc.cx(3, 1)
        qc.cx(3, 2)

        qc.cx(0, 1)
        qc.cx(0, 2)

        qc.cx(3, 1)
        qc.cx(3, 2)
        # qc.measure_all()

        pm = generate_preset_pass_manager(optimization_level=3, seed_transpiler=42)
        pm.init += StarPreRouting()

        result = pm.run(qc)

        self.assertTrue(Operator.from_circuit(result).equiv(Operator(qc)))

    def test_double_ghz(self):
        qc = QuantumCircuit(10)
        qc.h(0)
        qc.cx(0, range(1, 5))
        qc.h(9)
        qc.cx(9, range(8, 4, -1))

        pm = generate_preset_pass_manager(optimization_level=3, seed_transpiler=42)
        pm.init += StarPreRouting()
        result = pm.run(qc)

        self.assertEqual(Operator.from_circuit(result), Operator(qc))

    def test_linear_ghz_no_change(self):
        qc = QuantumCircuit(6)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(1, 2)
        qc.cx(2, 3)
        qc.cx(3, 4)
        qc.cx(4, 5)

        pm = generate_preset_pass_manager(optimization_level=3, seed_transpiler=42)
        pm.init += StarPreRouting()

        result = pm.run(qc)

        self.assertEqual(Operator.from_circuit(result), Operator(qc))

    def test_no_star(self):
        qc = QuantumCircuit(6)
        qc.h(0)
        qc.cx(0, 1)
        qc.cx(3, 2)
        qc.cx(0, 3)
        qc.cx(0, 4)
        qc.cx(1, 4)
        qc.cx(2, 1)

        pm = generate_preset_pass_manager(optimization_level=3, seed_transpiler=42)
        pm.init += StarPreRouting()
        result = pm.run(qc)

        self.assertTrue(Operator.from_circuit(result).equiv(qc))

    def test_10q_bv(self):
        num_qubits = 10
        qc = QuantumCircuit(num_qubits, num_qubits - 1)
        qc.x(num_qubits - 1)
        qc.h(qc.qubits)
        for i in range(num_qubits - 1):
            qc.cx(i, num_qubits - 1)
        qc.barrier()
        qc.h(qc.qubits[:-1])
        for i in range(num_qubits - 1):
            qc.measure(i, i)
        result = StarPreRouting()(qc)

        expected = QuantumCircuit(num_qubits, num_qubits - 1)
        expected.h(0)
        expected.h(1)
        expected.h(2)
        expected.h(3)
        expected.h(4)
        expected.h(5)
        expected.h(6)
        expected.h(7)
        expected.h(8)
        expected.x(9)
        expected.h(9)
        expected.cx(0, 9)
        expected.cx(1, 9)
        expected.swap(1, 9)
        expected.cx(2, 1)
        expected.swap(2, 1)
        expected.cx(3, 2)
        expected.swap(3, 2)
        expected.cx(4, 3)
        expected.swap(4, 3)
        expected.cx(5, 4)
        expected.swap(5, 4)
        expected.cx(6, 5)
        expected.swap(6, 5)
        expected.cx(7, 6)
        expected.swap(7, 6)
        expected.cx(8, 7)
        expected.barrier()
        expected.h(0)
        expected.h(1)
        expected.h(2)
        expected.h(3)
        expected.h(4)
        expected.h(5)
        expected.h(6)
        expected.h(8)
        expected.h(9)
        expected.measure(0, 0)
        expected.measure(9, 1)
        expected.measure(1, 2)
        expected.measure(2, 3)
        expected.measure(3, 4)
        expected.measure(4, 5)
        expected.measure(5, 6)
        expected.measure(6, 7)
        expected.measure(8, 8)
        self.assertEqual(result, expected)

    # Skip level 3 because of unitary synth introducing non-clifford gates
    @unittest.skipUnless(HAS_AER, "Aer required for clifford simulation")
    @ddt.data(0, 1)
    def test_100q_grid_full_path(self, opt_level):
        from qiskit_aer import AerSimulator

        num_qubits = 100
        coupling_map = CouplingMap.from_grid(10, 10)
        qc = QuantumCircuit(num_qubits, num_qubits - 1)
        qc.x(num_qubits - 1)
        qc.h(qc.qubits)
        for i in range(num_qubits - 1):
            qc.cx(i, num_qubits - 1)
        qc.barrier()
        qc.h(qc.qubits[:-1])
        for i in range(num_qubits - 1):
            qc.measure(i, i)
        pm = generate_preset_pass_manager(
            opt_level, basis_gates=["h", "cx", "x"], coupling_map=coupling_map
        )
        pm.init += StarPreRouting()
        result = pm.run(qc)
        counts_before = AerSimulator().run(qc).result().get_counts()
        counts_after = AerSimulator().run(result).result().get_counts()
        self.assertEqual(counts_before, counts_after)

    def test_10q_bv_no_barrier(self):
        num_qubits = 6
        qc = QuantumCircuit(num_qubits, num_qubits - 1)
        qc.x(num_qubits - 1)
        qc.h(qc.qubits)
        for i in range(num_qubits - 1):
            qc.cx(i, num_qubits - 1)
        qc.h(qc.qubits[:-1])

        pm = generate_preset_pass_manager(optimization_level=3, seed_transpiler=42)
        pm.init += StarPreRouting()

        result = pm.run(qc)
        self.assertTrue(Operator.from_circuit(result).equiv(Operator(qc)))

    # Skip level 3 because of unitary synth introducing non-clifford gates
    @unittest.skipUnless(HAS_AER, "Aer required for clifford simulation")
    @ddt.data(0, 1)
    def test_100q_grid_full_path_no_barrier(self, opt_level):
        from qiskit_aer import AerSimulator

        num_qubits = 100
        coupling_map = CouplingMap.from_grid(10, 10)
        qc = QuantumCircuit(num_qubits, num_qubits - 1)
        qc.x(num_qubits - 1)
        qc.h(qc.qubits)
        for i in range(num_qubits - 1):
            qc.cx(i, num_qubits - 1)
        qc.h(qc.qubits[:-1])
        for i in range(num_qubits - 1):
            qc.measure(i, i)
        pm = generate_preset_pass_manager(
            opt_level, basis_gates=["h", "cx", "x"], coupling_map=coupling_map
        )
        pm.init += StarPreRouting()
        result = pm.run(qc)
        counts_before = AerSimulator().run(qc).result().get_counts()
        counts_after = AerSimulator().run(result).result().get_counts()
        self.assertEqual(counts_before, counts_after)

    def test_hadamard_ordering(self):
        qc = QuantumCircuit(5)
        qc.h(0)
        qc.cx(0, 1)
        qc.h(0)
        qc.cx(0, 2)
        qc.h(0)
        qc.cx(0, 3)
        qc.h(0)
        qc.cx(0, 4)
        result = StarPreRouting()(qc)
        expected = QuantumCircuit(5)
        expected.h(0)
        expected.cx(0, 1)
        expected.h(0)
        expected.cx(0, 2)
        expected.swap(0, 2)
        expected.h(2)
        expected.cx(2, 3)
        expected.swap(2, 3)
        expected.h(3)
        expected.cx(3, 4)
        # expected.swap(3, 4)
        self.assertEqual(expected, result)

    def test_count_1_stars_starting_center(self):
        qc = QuantumCircuit(6)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(0, 3)
        qc.cx(0, 4)
        qc.cx(0, 5)
        spr = StarPreRouting()

        star_blocks, _ = spr.determine_star_blocks_processing(circuit_to_dag(qc), min_block_size=2)
        self.assertEqual(len(star_blocks), 1)
        self.assertEqual(len(star_blocks[0].nodes), 5)

    def test_count_1_stars_starting_branch(self):
        qc = QuantumCircuit(6)
        qc.cx(1, 0)
        qc.cx(2, 0)
        qc.cx(0, 3)
        qc.cx(0, 4)
        qc.cx(0, 5)
        spr = StarPreRouting()
        _ = spr(qc)

        star_blocks, _ = spr.determine_star_blocks_processing(circuit_to_dag(qc), min_block_size=2)
        self.assertEqual(len(star_blocks), 1)
        self.assertEqual(len(star_blocks[0].nodes), 5)

    def test_count_2_stars(self):
        qc = QuantumCircuit(6)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(0, 3)
        qc.cx(0, 4)
        qc.cx(0, 5)

        qc.cx(1, 2)
        qc.cx(1, 3)
        qc.cx(1, 4)
        qc.cx(1, 5)
        spr = StarPreRouting()
        _ = spr(qc)

        star_blocks, _ = spr.determine_star_blocks_processing(circuit_to_dag(qc), min_block_size=2)
        self.assertEqual(len(star_blocks), 2)
        self.assertEqual(len(star_blocks[0].nodes), 5)
        self.assertEqual(len(star_blocks[1].nodes), 4)

    def test_count_3_stars(self):
        qc = QuantumCircuit(6)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(0, 3)
        qc.cx(0, 4)
        qc.cx(0, 5)

        qc.cx(1, 2)
        qc.cx(1, 3)
        qc.cx(1, 4)
        qc.cx(1, 5)

        qc.cx(2, 3)
        qc.cx(2, 4)
        qc.cx(2, 5)
        spr = StarPreRouting()
        star_blocks, _ = spr.determine_star_blocks_processing(circuit_to_dag(qc), min_block_size=2)

        self.assertEqual(len(star_blocks), 3)
        self.assertEqual(len(star_blocks[0].nodes), 5)
        self.assertEqual(len(star_blocks[1].nodes), 4)
        self.assertEqual(len(star_blocks[2].nodes), 3)

    def test_count_70_qft_stars(self):
        qft_module = QFT(10, do_swaps=False).decompose()
        qftqc = QuantumCircuit(100)
        for i in range(10):
            qftqc.compose(qft_module, qubits=range(i * 10, (i + 1) * 10), inplace=True)
        spr = StarPreRouting()
        star_blocks, _ = spr.determine_star_blocks_processing(
            circuit_to_dag(qftqc), min_block_size=2
        )

        self.assertEqual(len(star_blocks), 80)
        star_len_list = [len([n for n in b.nodes if len(n.qargs) > 1]) for b in star_blocks]
        expected_star_size = {2, 3, 4, 5, 6, 7, 8, 9}
        self.assertEqual(set(star_len_list), expected_star_size)
        for i in expected_star_size:
            self.assertEqual(star_len_list.count(i), 10)

    def test_count_50_qft_stars(self):
        qft_module = QFT(10, do_swaps=False).decompose()
        qftqc = QuantumCircuit(10)
        for _ in range(10):
            qftqc.compose(qft_module, qubits=range(10), inplace=True)
        spr = StarPreRouting()
        _ = spr(qftqc)

        star_blocks, _ = spr.determine_star_blocks_processing(
            circuit_to_dag(qftqc), min_block_size=2
        )
        self.assertEqual(len(star_blocks), 50)
        star_len_list = [len([n for n in b.nodes if len(n.qargs) > 1]) for b in star_blocks]
        expected_star_size = {9}
        self.assertEqual(set(star_len_list), expected_star_size)

    def test_two_star_routing(self):
        qc = QuantumCircuit(4)
        qc.cx(0, 1)
        qc.cx(0, 2)

        qc.cx(2, 3)
        qc.cx(2, 1)

        spr = StarPreRouting()
        res = spr(qc)

        self.assertTrue(Operator.from_circuit(res).equiv(qc))

    def test_detect_two_opposite_stars_barrier(self):
        qc = QuantumCircuit(6)
        qc.cx(0, 1)
        qc.cx(0, 2)
        qc.cx(0, 3)
        qc.cx(0, 4)
        qc.barrier()
        qc.cx(5, 1)
        qc.cx(5, 2)
        qc.cx(5, 3)
        qc.cx(5, 4)

        spr = StarPreRouting()
        star_blocks, _ = spr.determine_star_blocks_processing(circuit_to_dag(qc), min_block_size=2)
        self.assertEqual(len(star_blocks), 2)
        self.assertEqual(len(star_blocks[0].nodes), 4)
        self.assertEqual(len(star_blocks[1].nodes), 4)

    def test_routing_after_star_prerouting(self):
        nq = 6
        qc = QFT(nq, do_swaps=False, insert_barriers=True).decompose()
        cm = CouplingMap.from_line(nq)

        pm_preroute = PassManager()
        pm_preroute.append(StarPreRouting())
        pm_preroute.append(VF2Layout(coupling_map=cm, seed=17))
        pm_preroute.append(ApplyLayout())
        pm_preroute.append(SabreSwap(coupling_map=cm, seed=17))

        pm_sabre = PassManager()
        pm_sabre.append(SabreLayout(coupling_map=cm, seed=17))

        res_sabre = pm_sabre.run(qc)
        res_star = pm_sabre.run(qc)

        self.assertTrue(Operator.from_circuit(res_sabre), qc)
        self.assertTrue(Operator.from_circuit(res_star), qc)
        self.assertTrue(Operator.from_circuit(res_star), Operator.from_circuit(res_sabre))

    @ddt.data(4, 8, 16, 32)
    def test_qft_linearization(self, num_qubits):
        """Test the QFT circuit to verify if it is linearized and requires n-2 swaps."""

        qc = QFT(num_qubits, do_swaps=False, insert_barriers=True).decompose()
        dag = circuit_to_dag(qc)
        new_dag = StarPreRouting().run(dag)
        new_qc = dag_to_circuit(new_dag)

        # Check that resulting result has n-2 swaps, where n is the number of cp gates
        swap_count = new_qc.count_ops().get("swap", 0)
        cp_count = new_qc.count_ops().get("cp", 0)
        self.assertEqual(swap_count, cp_count - 2)

        # Confirm linearization by checking that the number of edges is equal to the number of nodes
        interaction_graph = build_interaction_graph(new_dag, strict_direction=False)[0]
        num_edges = interaction_graph.num_edges()
        num_nodes = interaction_graph.num_nodes()
        self.assertEqual(num_edges, num_nodes - 1)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

# pylint: disable=unused-argument

"""Chart axis generators.

A collection of functions that generate drawings from formatted input data.
See py:mod:`qiskit.visualization.pulse_v2.types` for more info on the required data.

In this module the input data is `types.ChartAxis`.

An end-user can write arbitrary functions that generate custom drawings.
Generators in this module are called with the `formatter` and `device` kwargs.
These data provides stylesheet configuration and backend system configuration.

The format of generator is restricted to:

    ```python

    def my_object_generator(data: ChartAxis,
                            formatter: Dict[str, Any],
                            device: DrawerBackendInfo) -> List[ElementaryData]:
        pass
    ```

Arbitrary generator function satisfying the above format can be accepted.
Returned `ElementaryData` can be arbitrary subclasses that are implemented in
the plotter API.
"""
from typing import Dict, Any, List

from qiskit.visualization.pulse_v2 import drawings, types, device_info


def gen_baseline(
    data: types.ChartAxis, formatter: Dict[str, Any], device: device_info.DrawerBackendInfo
) -> List[drawings.LineData]:
    """Generate the baseline associated with the chart.

    Stylesheets:
        - The `baseline` style is applied.

    Args:
        data: Chart axis data to draw.
        formatter: Dictionary of stylesheet settings.
        device: Backend configuration.

    Returns:
        List of `LineData` drawings.
    """
    style = {
        "alpha": formatter["alpha.baseline"],
        "zorder": formatter["layer.baseline"],
        "linewidth": formatter["line_width.baseline"],
        "linestyle": formatter["line_style.baseline"],
        "color": formatter["color.baseline"],
    }

    baseline = drawings.LineData(
        data_type=types.LineType.BASELINE,
        channels=data.channels,
        xvals=[types.AbstractCoordinate.LEFT, types.AbstractCoordinate.RIGHT],
        yvals=[0, 0],
        ignore_scaling=True,
        styles=style,
    )

    return [baseline]


def gen_chart_name(
    data: types.ChartAxis, formatter: Dict[str, Any], device: device_info.DrawerBackendInfo
) -> List[drawings.TextData]:
    """Generate the name of chart.

    Stylesheets:
        - The `axis_label` style is applied.

    Args:
        data: Chart axis data to draw.
        formatter: Dictionary of stylesheet settings.
        device: Backend configuration.

    Returns:
        List of `TextData` drawings.
    """
    style = {
        "zorder": formatter["layer.axis_label"],
        "color": formatter["color.axis_label"],
        "size": formatter["text_size.axis_label"],
        "va": "center",
        "ha": "right",
    }

    text = drawings.TextData(
        data_type=types.LabelType.CH_NAME,
        channels=data.channels,
        xvals=[types.AbstractCoordinate.LEFT],
        yvals=[0],
        text=data.name,
        ignore_scaling=True,
        styles=style,
    )

    return [text]


def gen_chart_scale(
    data: types.ChartAxis, formatter: Dict[str, Any], device: device_info.DrawerBackendInfo
) -> List[drawings.TextData]:
    """Generate the current scaling value of the chart.

    Stylesheets:
        - The `axis_label` style is applied.
        - The `annotate` style is partially applied for the font size.

    Args:
        data: Chart axis data to draw.
        formatter: Dictionary of stylesheet settings.
        device: Backend configuration.

    Returns:
        List of `TextData` drawings.
    """
    style = {
        "zorder": formatter["layer.axis_label"],
        "color": formatter["color.axis_label"],
        "size": formatter["text_size.annotate"],
        "va": "center",
        "ha": "right",
    }

    scale_val = f"x{types.DynamicString.SCALE}"

    text = drawings.TextData(
        data_type=types.LabelType.CH_INFO,
        channels=data.channels,
        xvals=[types.AbstractCoordinate.LEFT],
        yvals=[-formatter["label_offset.chart_info"]],
        text=scale_val,
        ignore_scaling=True,
        styles=style,
    )

    return [text]


def gen_channel_freqs(
    data: types.ChartAxis, formatter: Dict[str, Any], device: device_info.DrawerBackendInfo
) -> List[drawings.TextData]:
    """Generate the frequency values of associated channels.

    Stylesheets:
        - The `axis_label` style is applied.
        - The `annotate` style is partially applied for the font size.

    Args:
        data: Chart axis data to draw.
        formatter: Dictionary of stylesheet settings.
        device: Backend configuration.

    Returns:
        List of `TextData` drawings.
    """
    style = {
        "zorder": formatter["layer.axis_label"],
        "color": formatter["color.axis_label"],
        "size": formatter["text_size.annotate"],
        "va": "center",
        "ha": "right",
    }

    if len(data.channels) > 1:
        sources = []
        for chan in data.channels:
            freq = device.get_channel_frequency(chan)
            if not freq:
                continue
            sources.append(f"{chan.name.upper()}: {freq / 1e9:.2f} GHz")
        freq_text = ", ".join(sources)
    else:
        freq = device.get_channel_frequency(data.channels[0])
        if freq:
            freq_text = f"{freq / 1e9:.2f} GHz"
        else:
            freq_text = ""

    text = drawings.TextData(
        data_type=types.LabelType.CH_INFO,
        channels=data.channels,
        xvals=[types.AbstractCoordinate.LEFT],
        yvals=[-formatter["label_offset.chart_info"]],
        text=freq_text or "no freq.",
        ignore_scaling=True,
        styles=style,
    )

    return [text]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Check if all gates in the DAGCircuit are in the specified basis gates."""

from qiskit.transpiler.basepasses import AnalysisPass

from qiskit._accelerate.gates_in_basis import (
    any_gate_missing_from_basis,
    any_gate_missing_from_target,
)


class GatesInBasis(AnalysisPass):
    """Check if all gates in a DAG are in a given set of gates"""

    def __init__(self, basis_gates=None, target=None):
        """Initialize the GatesInBasis pass.

        Args:
            basis_gates (list): The list of strings representing the set of basis gates.
            target (Target): The target representing the backend. If specified
                this will be used instead of the ``basis_gates`` parameter
        """
        super().__init__()
        self._basis_gates = None
        if basis_gates is not None:
            self._basis_gates = set(basis_gates).union(
                {"measure", "reset", "barrier", "snapshot", "delay", "store"}
            )
        self._target = target

    def run(self, dag):
        """Run the GatesInBasis pass on `dag`."""
        if self._basis_gates is None and self._target is None:
            self.property_set["all_gates_in_basis"] = True
            return
        if self._target is not None:
            gates_out_of_basis = any_gate_missing_from_target(dag, self._target)
        else:
            gates_out_of_basis = any_gate_missing_from_basis(dag, self._basis_gates)
        self.property_set["all_gates_in_basis"] = not gates_out_of_basis



# This code is part of Qiskit.
#
# (C) Copyright IBM 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
.. currentmodule:: qiskit.utils.optionals

Qiskit has several features that are enabled only if certain *optional* dependencies
are satisfied. This module, :mod:`qiskit.utils.optionals`, has a collection of objects that can
be used to test if certain functionality is available, and optionally raise
:class:`.MissingOptionalLibraryError` if the functionality is not available.


Available Testers
=================

Qiskit Components
-----------------

.. py:data:: HAS_AER

    `Qiskit Aer <https://qiskit.github.io/qiskit-aer/>` provides high-performance simulators for
    the quantum circuits constructed within Qiskit.

.. py:data:: HAS_IBMQ

    The :mod:`Qiskit IBMQ Provider <qiskit.providers.ibmq>` is used for accessing IBM Quantum
    hardware in the IBM cloud.

.. py:data:: HAS_IGNIS

    :mod:`Qiskit Ignis <qiskit.ignis>` provides tools for quantum hardware verification, noise
    characterization, and error correction.

.. py:data:: HAS_TOQM

    `Qiskit TOQM <https://github.com/qiskit-toqm/qiskit-toqm>`__ provides transpiler passes
    for the `Time-optimal Qubit mapping algorithm <https://doi.org/10.1145/3445814.3446706>`__.


External Python Libraries
-------------------------

.. py:data:: HAS_CONSTRAINT

    `python-constraint <https://github.com/python-constraint/python-constraint>`__ is a
    constraint satisfaction problem solver, used in the :class:`~.CSPLayout` transpiler pass.

.. py:data:: HAS_CPLEX

    The `IBM CPLEX Optimizer <https://www.ibm.com/analytics/cplex-optimizer>`__ is a
    high-performance mathematical programming solver for linear, mixed-integer and quadratic
    programming. This is no longer by Qiskit, but it weas historically and the optional
    remains for backwards compatibility.

.. py:data:: HAS_CVXPY

    `CVXPY <https://www.cvxpy.org/>`__ is a Python package for solving convex optimization
    problems.  It is required for calculating diamond norms with
    :func:`.quantum_info.diamond_norm`.

.. py:data:: HAS_DOCPLEX

    `IBM Decision Optimization CPLEX Modelling
    <http://ibmdecisionoptimization.github.io/docplex-doc/>`__ is a library for prescriptive
    analysis.  Like CPLEX, this is no longer by Qiskit, but it weas historically and the
    optional remains for backwards compatibility.

.. py:data:: HAS_FIXTURES

    The test suite has additional features that are available if the optional `fixtures
    <https://launchpad.net/python-fixtures>`__ module is installed.  This generally also needs
    :data:`HAS_TESTTOOLS` as well.  This is generally only needed for Qiskit developers.

.. py:data:: HAS_IPYTHON

    If `the IPython kernel <https://ipython.org/>`__ is available, certain additional
    visualizations and line magics are made available.

.. py:data:: HAS_IPYWIDGETS

    Monitoring widgets for jobs running on external backends can be provided if `ipywidgets
    <https://ipywidgets.readthedocs.io/en/latest/>`__ is available.

.. py:data:: HAS_JAX

    Some methods of gradient calculation within :mod:`.opflow.gradients` require `JAX
    <https://github.com/google/jax>`__ for autodifferentiation.

.. py:data:: HAS_JUPYTER

    Some of the tests require a complete `Jupyter <https://jupyter.org/>`__ installation to test
    interactivity features.

.. py:data:: HAS_MATPLOTLIB

    Qiskit provides several visualization tools in the :mod:`.visualization` module.
    Almost all of these are built using `Matplotlib <https://matplotlib.org/>`__, which must
    be installed in order to use them.

.. py:data:: HAS_NETWORKX

    No longer used by Qiskit.  Internally, Qiskit now uses the high-performance `rustworkx
    <https://github.com/Qiskit/rustworkx>`__ library as a core dependency, and during the
    change-over period, it was sometimes convenient to convert things into the Python-only
    `NetworkX <https://networkx.org/>`__ format.  Some tests of application modules, such as
    `Qiskit Nature <https://qiskit-community.github.io/qiskit-nature/>`__ still use NetworkX.

.. py:data:: HAS_NLOPT

    `NLopt <https://nlopt.readthedocs.io/en/latest/>`__ is a nonlinear optimization library,
    used by the global optimizers in the :mod:`.algorithms.optimizers` module.

.. py:data:: HAS_PIL

    PIL is a Python image-manipulation library.  Qiskit actually uses the `pillow
    <https://pillow.readthedocs.io/en/stable/>`__ fork of PIL if it is available when generating
    certain visualizations, for example of both :class:`.QuantumCircuit` and
    :class:`.DAGCircuit` in certain modes.

.. py:data:: HAS_PYDOT

    For some graph visualizations, Qiskit uses `pydot <https://github.com/pydot/pydot>`__ as an
    interface to GraphViz (see :data:`HAS_GRAPHVIZ`).

.. py:data:: HAS_PYGMENTS

    Pygments is a code highlighter and formatter used by many environments that involve rich
    display of code blocks, including Sphinx and Jupyter.  Qiskit uses this when producing rich
    output for these environments.

.. py:data:: HAS_PYLATEX

    Various LaTeX-based visualizations, especially the circuit drawers, need access to the
    `pylatexenc <https://github.com/phfaist/pylatexenc>`__ project to work correctly.

.. py:data:: HAS_QASM3_IMPORT

    The functions :func:`.qasm3.load` and :func:`.qasm3.loads` for importing OpenQASM 3 programs
    into :class:`.QuantumCircuit` instances use `an external importer package
    <https://qiskit.github.io/qiskit-qasm3-import>`__.

.. py:data:: HAS_SEABORN

    Qiskit provides several visualization tools in the :mod:`.visualization` module.  Some
    of these are built using `Seaborn <https://seaborn.pydata.org/>`__, which must be installed
    in order to use them.

.. py:data:: HAS_SKLEARN

    Some of the gradient functions in :mod:`.opflow.gradients` use regularisation methods from
    `Scikit Learn <https://scikit-learn.org/stable/>`__.

.. py:data:: HAS_SKQUANT

    Some of the optimisers in :mod:`.algorithms.optimizers` are based on those found in `Scikit
    Quant <https://github.com/scikit-quant/scikit-quant>`__, which must be installed to use
    them.

.. py:data:: HAS_SQSNOBFIT

    `SQSnobFit <https://pypi.org/project/SQSnobFit/>`__ is a library for the "stable noisy
    optimization by branch and fit" algorithm.  It is used by the :class:`.SNOBFIT` optimizer.

.. py:data:: HAS_SYMENGINE

    `Symengine <https://github.com/symengine/symengine>`__ is a fast C++ backend for the
    symbolic-manipulation library `Sympy <https://www.sympy.org/en/index.html>`__.  Qiskit uses
    special methods from Symengine to accelerate its handling of
    :class:`~.circuit.Parameter`\\ s if available.

.. py:data:: HAS_TESTTOOLS

    Qiskit's test suite has more advanced functionality available if the optional
    `testtools <https://pypi.org/project/testtools/>`__ library is installed.  This is generally
    only needed for Qiskit developers.

.. py:data:: HAS_TWEEDLEDUM

    `Tweedledum <https://github.com/boschmitt/tweedledum>`__ is an extension library for
    synthesis and optimization of circuits that may involve classical oracles.  Qiskit's
    :class:`.PhaseOracle` uses this, which is used in turn by amplification algorithms via
    the :class:`.AmplificationProblem`.

.. py:data:: HAS_Z3

    `Z3 <https://github.com/Z3Prover/z3>`__ is a theorem prover, used in the
    :class:`.CrosstalkAdaptiveSchedule` and :class:`.HoareOptimizer` transpiler passes.

External Command-Line Tools
---------------------------

.. py:data:: HAS_GRAPHVIZ

    For some graph visualizations, Qiskit uses the `GraphViz <https://graphviz.org/>`__
    visualization tool via its ``pydot`` interface (see :data:`HAS_PYDOT`).

.. py:data:: HAS_PDFLATEX

    Visualization tools that use LaTeX in their output, such as the circuit drawers, require
    ``pdflatex`` to be available.  You will generally need to ensure that you have a working
    LaTeX installation available, and the ``qcircuit.tex`` package.

.. py:data:: HAS_PDFTOCAIRO

    Visualization tools that convert LaTeX-generated files into rasterized images use the
    ``pdftocairo`` tool.  This is part of the `Poppler suite of PDF tools
    <https://poppler.freedesktop.org/>`__.


Lazy Checker Classes
====================

.. currentmodule:: qiskit.utils

Each of the lazy checkers is an instance of :class:`.LazyDependencyManager` in one of its two
subclasses: :class:`.LazyImportTester` and :class:`.LazySubprocessTester`.  These should be imported
from :mod:`.utils` directly if required, such as::

    from qiskit.utils import LazyImportTester

.. autoclass:: qiskit.utils.LazyDependencyManager
    :members:

.. autoclass:: qiskit.utils.LazyImportTester
.. autoclass:: qiskit.utils.LazySubprocessTester
"""

# NOTE: If you're changing this file, sync it with `requirements-optional.txt` and potentially
# `pyproject.toml` as well.

import logging as _logging

from .lazy_tester import (
    LazyImportTester as _LazyImportTester,
    LazySubprocessTester as _LazySubprocessTester,
)

_logger = _logging.getLogger(__name__)

HAS_AER = _LazyImportTester(
    "qiskit_aer",
    name="Qiskit Aer",
    install="pip install qiskit-aer",
)
HAS_IBMQ = _LazyImportTester(
    "qiskit.providers.ibmq",
    name="IBMQ Provider",
    install="pip install qiskit-ibmq-provider",
)
HAS_IGNIS = _LazyImportTester(
    "qiskit.ignis",
    name="Qiskit Ignis",
    install="pip install qiskit-ignis",
)
HAS_TOQM = _LazyImportTester("qiskit_toqm", name="Qiskit TOQM", install="pip install qiskit-toqm")

HAS_CONSTRAINT = _LazyImportTester(
    "constraint",
    name="python-constraint",
    install="pip install python-constraint",
)

HAS_CPLEX = _LazyImportTester(
    "cplex",
    install="pip install cplex",
    msg="This may not be possible for all Python versions and OSes",
)
HAS_CVXPY = _LazyImportTester("cvxpy", install="pip install cvxpy")
HAS_DOCPLEX = _LazyImportTester(
    {"docplex": (), "docplex.mp.model": ("Model",)},
    install="pip install docplex",
    msg="This may not be possible for all Python versions and OSes",
)
HAS_FIXTURES = _LazyImportTester("fixtures", install="pip install fixtures")
HAS_IPYTHON = _LazyImportTester("IPython", install="pip install ipython")
HAS_IPYWIDGETS = _LazyImportTester("ipywidgets", install="pip install ipywidgets")
HAS_JAX = _LazyImportTester(
    {"jax": ("grad", "jit"), "jax.numpy": ()},
    name="jax",
    install="pip install jax",
)
HAS_JUPYTER = _LazyImportTester(["jupyter", "nbformat", "nbconvert"], install="pip install jupyter")
HAS_MATPLOTLIB = _LazyImportTester(
    ("matplotlib.patches", "matplotlib.pyplot"),
    name="matplotlib",
    install="pip install matplotlib",
)
HAS_NETWORKX = _LazyImportTester("networkx", install="pip install networkx")

HAS_NLOPT = _LazyImportTester("nlopt", name="NLopt Optimizer", install="pip install nlopt")
HAS_PIL = _LazyImportTester("PIL.Image", name="pillow", install="pip install pillow")
HAS_PYDOT = _LazyImportTester("pydot", install="pip install pydot")
HAS_PYGMENTS = _LazyImportTester("pygments", install="pip install pygments")
HAS_PYLATEX = _LazyImportTester(
    {
        "pylatexenc.latex2text": ("LatexNodes2Text",),
        "pylatexenc.latexencode": ("utf8tolatex",),
    },
    name="pylatexenc",
    install="pip install pylatexenc",
)
HAS_QASM3_IMPORT = _LazyImportTester(
    "qiskit_qasm3_import", install="pip install qiskit_qasm3_import"
)
HAS_SEABORN = _LazyImportTester("seaborn", install="pip install seaborn")
HAS_SKLEARN = _LazyImportTester(
    {"sklearn.linear_model": ("Ridge", "Lasso")},
    name="scikit-learn",
    install="pip install scikit-learn",
)
HAS_SKQUANT = _LazyImportTester(
    "skquant.opt",
    name="scikit-quant",
    install="pip install scikit-quant",
)
HAS_SQSNOBFIT = _LazyImportTester("SQSnobFit", install="pip install SQSnobFit")
HAS_SYMENGINE = _LazyImportTester("symengine", install="pip install symengine")
HAS_TESTTOOLS = _LazyImportTester("testtools", install="pip install testtools")
HAS_TWEEDLEDUM = _LazyImportTester("tweedledum", install="pip install tweedledum")
HAS_Z3 = _LazyImportTester("z3", install="pip install z3-solver")

HAS_GRAPHVIZ = _LazySubprocessTester(
    ("dot", "-V"),
    name="Graphviz",
    msg=(
        "To install, follow the instructions at https://graphviz.org/download/."
        " Qiskit needs the Graphviz binaries, which the 'graphviz' package on pip does not install."
        " You must install the actual Graphviz software"
    ),
)
HAS_PDFLATEX = _LazySubprocessTester(
    ("pdflatex", "-version"),
    msg="You will likely need to install a full LaTeX distribution for your system",
)
HAS_PDFTOCAIRO = _LazySubprocessTester(
    ("pdftocairo", "-v"),
    msg="This is part of the 'poppler' set of PDF utilities",
)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""Tests continuous pulse functions."""

import numpy as np

from qiskit.pulse.library import continuous
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestContinuousPulses(QiskitTestCase):
    """Test continuous pulses."""

    def test_constant(self):
        """Test constant pulse."""
        amp = 0.5j
        samples = 50
        times = np.linspace(0, 10, samples)

        constant_arr = continuous.constant(times, amp=amp)

        self.assertEqual(constant_arr.dtype, np.complex128)
        np.testing.assert_equal(constant_arr, amp)
        self.assertEqual(len(constant_arr), samples)

    def test_zero(self):
        """Test constant pulse."""
        times = np.linspace(0, 10, 50)
        zero_arr = continuous.zero(times)

        self.assertEqual(zero_arr.dtype, np.complex128)
        np.testing.assert_equal(zero_arr, 0.0)
        self.assertEqual(len(zero_arr), 50)

    def test_square(self):
        """Test square wave."""
        amp = 0.5
        freq = 0.2
        samples = 100
        times = np.linspace(0, 10, samples)
        square_arr = continuous.square(times, amp=amp, freq=freq)
        # with new phase
        square_arr_phased = continuous.square(times, amp=amp, freq=freq, phase=np.pi / 2)

        self.assertEqual(square_arr.dtype, np.complex128)

        self.assertAlmostEqual(square_arr[0], amp)
        # test constant
        self.assertAlmostEqual(square_arr[1] - square_arr[0], 0.0)
        self.assertAlmostEqual(square_arr[25], -amp)
        self.assertAlmostEqual(square_arr_phased[0], -amp)
        # Assert bounded between -amp and amp
        self.assertTrue(np.all((-amp <= square_arr) & (square_arr <= amp)))
        self.assertEqual(len(square_arr), samples)

    def test_sawtooth(self):
        """Test sawtooth wave."""
        amp = 0.5
        freq = 0.2
        samples = 101
        times, dt = np.linspace(0, 10, samples, retstep=True)
        sawtooth_arr = continuous.sawtooth(times, amp=amp, freq=freq)
        # with new phase
        sawtooth_arr_phased = continuous.sawtooth(times, amp=amp, freq=freq, phase=np.pi / 2)

        self.assertEqual(sawtooth_arr.dtype, np.complex128)

        self.assertAlmostEqual(sawtooth_arr[0], 0.0)
        # test slope
        self.assertAlmostEqual((sawtooth_arr[1] - sawtooth_arr[0]) / dt, 2 * amp * freq)
        self.assertAlmostEqual(sawtooth_arr[24], 0.48)
        self.assertAlmostEqual(sawtooth_arr[50], 0.0)
        self.assertAlmostEqual(sawtooth_arr[75], -amp)
        self.assertAlmostEqual(sawtooth_arr_phased[0], -amp)
        # Assert bounded between -amp and amp
        self.assertTrue(np.all((-amp <= sawtooth_arr) & (sawtooth_arr <= amp)))
        self.assertEqual(len(sawtooth_arr), samples)

    def test_triangle(self):
        """Test triangle wave."""
        amp = 0.5
        freq = 0.2
        samples = 101
        times, dt = np.linspace(0, 10, samples, retstep=True)
        triangle_arr = continuous.triangle(times, amp=amp, freq=freq)
        # with new phase
        triangle_arr_phased = continuous.triangle(times, amp=amp, freq=freq, phase=np.pi / 2)

        self.assertEqual(triangle_arr.dtype, np.complex128)

        self.assertAlmostEqual(triangle_arr[0], 0.0)
        # test slope
        self.assertAlmostEqual((triangle_arr[1] - triangle_arr[0]) / dt, 4 * amp * freq)
        self.assertAlmostEqual(triangle_arr[12], 0.48)
        self.assertAlmostEqual(triangle_arr[13], 0.48)
        self.assertAlmostEqual(triangle_arr[50], 0.0)
        self.assertAlmostEqual(triangle_arr_phased[0], amp)
        # Assert bounded between -amp and amp
        self.assertTrue(np.all((-amp <= triangle_arr) & (triangle_arr <= amp)))
        self.assertEqual(len(triangle_arr), samples)

    def test_cos(self):
        """Test cosine wave."""
        amp = 0.5
        period = 5
        freq = 1 / period
        samples = 101
        times = np.linspace(0, 10, samples)
        cos_arr = continuous.cos(times, amp=amp, freq=freq)
        # with new phase
        cos_arr_phased = continuous.cos(times, amp=amp, freq=freq, phase=np.pi / 2)

        self.assertEqual(cos_arr.dtype, np.complex128)

        # Assert starts at 1
        self.assertAlmostEqual(cos_arr[0], amp)
        self.assertAlmostEqual(cos_arr[6], 0.3644, places=2)
        self.assertAlmostEqual(cos_arr[25], -amp)
        self.assertAlmostEqual(cos_arr[50], amp)
        self.assertAlmostEqual(cos_arr_phased[0], 0.0)
        # Assert bounded between -amp and amp
        self.assertTrue(np.all((-amp <= cos_arr) & (cos_arr <= amp)))
        self.assertEqual(len(cos_arr), samples)

    def test_sin(self):
        """Test sine wave."""
        amp = 0.5
        period = 5
        freq = 1 / period
        samples = 101
        times = np.linspace(0, 10, samples)
        sin_arr = continuous.sin(times, amp=amp, freq=freq)
        # with new phase
        sin_arr_phased = continuous.sin(times, amp=0.5, freq=1 / 5, phase=np.pi / 2)

        self.assertEqual(sin_arr.dtype, np.complex128)

        # Assert starts at 1
        self.assertAlmostEqual(sin_arr[0], 0.0)
        self.assertAlmostEqual(sin_arr[6], 0.3427, places=2)
        self.assertAlmostEqual(sin_arr[25], 0.0)
        self.assertAlmostEqual(sin_arr[13], amp, places=2)
        self.assertAlmostEqual(sin_arr_phased[0], amp)
        # Assert bounded between -amp and amp
        self.assertTrue(np.all((-amp <= sin_arr) & (sin_arr <= amp)))
        self.assertEqual(len(sin_arr), samples)

    def test_gaussian(self):
        """Test gaussian pulse."""
        amp = 0.5
        duration = 20
        center = duration / 2
        sigma = 2
        times, dt = np.linspace(0, duration, 1001, retstep=True)
        gaussian_arr = continuous.gaussian(times, amp, center, sigma)
        gaussian_arr_zeroed = continuous.gaussian(
            np.array([-1, center, duration + 1]),
            amp,
            center,
            sigma,
            zeroed_width=2 * (center + 1),
            rescale_amp=True,
        )

        self.assertEqual(gaussian_arr.dtype, np.complex128)

        center_time = np.argmax(gaussian_arr)
        self.assertAlmostEqual(times[center_time], center)
        self.assertAlmostEqual(gaussian_arr[center_time], amp)
        self.assertAlmostEqual(gaussian_arr_zeroed[0], 0.0, places=6)
        self.assertAlmostEqual(gaussian_arr_zeroed[1], amp)
        self.assertAlmostEqual(gaussian_arr_zeroed[2], 0.0, places=6)
        self.assertAlmostEqual(
            np.sum(gaussian_arr * dt), amp * np.sqrt(2 * np.pi * sigma**2), places=3
        )

    def test_gaussian_deriv(self):
        """Test gaussian derivative pulse."""
        amp = 0.5
        center = 10
        sigma = 2
        times, dt = np.linspace(0, 20, 1000, retstep=True)
        deriv_prefactor = -(sigma**2) / (times - center)

        gaussian_deriv_arr = continuous.gaussian_deriv(times, amp, center, sigma)
        gaussian_arr = gaussian_deriv_arr * deriv_prefactor

        self.assertEqual(gaussian_deriv_arr.dtype, np.complex128)

        self.assertAlmostEqual(
            continuous.gaussian_deriv(np.array([0]), amp, center, sigma)[0], 0, places=5
        )
        self.assertAlmostEqual(
            np.sum(gaussian_arr * dt), amp * np.sqrt(2 * np.pi * sigma**2), places=3
        )

    def test_sech(self):
        """Test sech pulse."""
        amp = 0.5
        duration = 40
        center = duration / 2
        sigma = 2
        times, dt = np.linspace(0, duration, 1001, retstep=True)
        sech_arr = continuous.sech(times, amp, center, sigma)
        sech_arr_zeroed = continuous.sech(np.array([-1, center, duration + 1]), amp, center, sigma)

        self.assertEqual(sech_arr.dtype, np.complex128)

        center_time = np.argmax(sech_arr)
        self.assertAlmostEqual(times[center_time], center)
        self.assertAlmostEqual(sech_arr[center_time], amp)
        self.assertAlmostEqual(sech_arr_zeroed[0], 0.0, places=2)
        self.assertAlmostEqual(sech_arr_zeroed[1], amp)
        self.assertAlmostEqual(sech_arr_zeroed[2], 0.0, places=2)
        self.assertAlmostEqual(np.sum(sech_arr * dt), amp * np.pi * sigma, places=3)

    def test_sech_deriv(self):
        """Test sech derivative pulse."""
        amp = 0.5
        center = 20
        sigma = 2
        times = np.linspace(0, 40, 1000)

        sech_deriv_arr = continuous.sech_deriv(times, amp, center, sigma)

        self.assertEqual(sech_deriv_arr.dtype, np.complex128)

        self.assertAlmostEqual(
            continuous.sech_deriv(np.array([0]), amp, center, sigma)[0], 0, places=3
        )

    def test_gaussian_square(self):
        """Test gaussian square pulse."""
        amp = 0.5
        center = 10
        width = 2
        sigma = 0.1
        times, dt = np.linspace(0, 20, 2001, retstep=True)
        gaussian_square_arr = continuous.gaussian_square(times, amp, center, width, sigma)

        self.assertEqual(gaussian_square_arr.dtype, np.complex128)

        self.assertEqual(gaussian_square_arr[1000], amp)
        # test half gaussian rise/fall
        self.assertAlmostEqual(
            np.sum(gaussian_square_arr[:900] * dt) * 2,
            amp * np.sqrt(2 * np.pi * sigma**2),
            places=2,
        )
        self.assertAlmostEqual(
            np.sum(gaussian_square_arr[1100:] * dt) * 2,
            amp * np.sqrt(2 * np.pi * sigma**2),
            places=2,
        )
        # test for continuity at gaussian/square boundaries
        gauss_rise_end_time = center - width / 2
        gauss_fall_start_time = center + width / 2
        epsilon = 0.01
        rise_times, dt_rise = np.linspace(
            gauss_rise_end_time - epsilon, gauss_rise_end_time + epsilon, 1001, retstep=True
        )
        fall_times, dt_fall = np.linspace(
            gauss_fall_start_time - epsilon, gauss_fall_start_time + epsilon, 1001, retstep=True
        )
        gaussian_square_rise_arr = continuous.gaussian_square(rise_times, amp, center, width, sigma)
        gaussian_square_fall_arr = continuous.gaussian_square(fall_times, amp, center, width, sigma)

        # should be locally approximated by amp*dt^2/(2*sigma^2)
        self.assertAlmostEqual(
            amp * dt_rise**2 / (2 * sigma**2),
            gaussian_square_rise_arr[500] - gaussian_square_rise_arr[499],
        )
        self.assertAlmostEqual(
            amp * dt_fall**2 / (2 * sigma**2),
            gaussian_square_fall_arr[501] - gaussian_square_fall_arr[500],
        )

    def test_drag(self):
        """Test drag pulse."""
        amp = 0.5
        center = 10
        sigma = 0.1
        beta = 0
        times = np.linspace(0, 20, 2001)
        # test that we recover gaussian for beta=0
        gaussian_arr = continuous.gaussian(
            times, amp, center, sigma, zeroed_width=2 * (center + 1), rescale_amp=True
        )

        drag_arr = continuous.drag(
            times, amp, center, sigma, beta=beta, zeroed_width=2 * (center + 1), rescale_amp=True
        )

        self.assertEqual(drag_arr.dtype, np.complex128)

        np.testing.assert_equal(drag_arr, gaussian_arr)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
ParameterExpression Class to enable creating simple expressions of Parameters.
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import IntEnum
from typing import Callable, Union

import numbers
import operator

import numpy
import symengine

from qiskit.circuit.exceptions import CircuitError

# This type is redefined at the bottom to insert the full reference to "ParameterExpression", so it
# can safely be used by runtime type-checkers like Sphinx.  Mypy does not need this because it
# handles the references by static analysis.
ParameterValueType = Union["ParameterExpression", float]


class _OPCode(IntEnum):
    ADD = 0
    SUB = 1
    MUL = 2
    DIV = 3
    POW = 4
    SIN = 5
    COS = 6
    TAN = 7
    ASIN = 8
    ACOS = 9
    EXP = 10
    LOG = 11
    SIGN = 12
    GRAD = 13
    CONJ = 14
    SUBSTITUTE = 15
    ABS = 16
    ATAN = 17
    RSUB = 18
    RDIV = 19
    RPOW = 20


_OP_CODE_MAP = (
    "__add__",
    "__sub__",
    "__mul__",
    "__truediv__",
    "__pow__",
    "sin",
    "cos",
    "tan",
    "arcsin",
    "arccos",
    "exp",
    "log",
    "sign",
    "gradient",
    "conjugate",
    "subs",
    "abs",
    "arctan",
    "__rsub__",
    "__rtruediv__",
    "__rpow__",
)


def op_code_to_method(op_code: _OPCode):
    """Return the method name for a given op_code."""
    return _OP_CODE_MAP[op_code]


@dataclass
class _INSTRUCTION:
    op: _OPCode
    lhs: ParameterValueType | None
    rhs: ParameterValueType | None = None


@dataclass
class _SUBS:
    binds: dict
    op: _OPCode = _OPCode.SUBSTITUTE


class ParameterExpression:
    """ParameterExpression class to enable creating expressions of Parameters."""

    __slots__ = [
        "_parameter_symbols",
        "_parameter_keys",
        "_symbol_expr",
        "_name_map",
        "_qpy_replay",
        "_standalone_param",
    ]

    def __init__(self, symbol_map: dict, expr, *, _qpy_replay=None):
        """Create a new :class:`ParameterExpression`.

        Not intended to be called directly, but to be instantiated via operations
        on other :class:`Parameter` or :class:`ParameterExpression` objects.

        Args:
            symbol_map (Dict[Parameter, [ParameterExpression, float, or int]]):
                Mapping of :class:`Parameter` instances to the :class:`sympy.Symbol`
                serving as their placeholder in expr.
            expr (sympy.Expr): Expression of :class:`sympy.Symbol` s.
        """
        # NOTE: `Parameter.__init__` does not call up to this method, since this method is dependent
        # on `Parameter` instances already being initialized enough to be hashable.  If changing
        # this method, check that `Parameter.__init__` and `__setstate__` are still valid.
        self._parameter_symbols = symbol_map
        self._parameter_keys = frozenset(p._hash_key() for p in self._parameter_symbols)
        self._symbol_expr = expr
        self._name_map: dict | None = None
        self._standalone_param = False
        if _qpy_replay is not None:
            self._qpy_replay = _qpy_replay
        else:
            self._qpy_replay = []

    @property
    def parameters(self) -> set:
        """Returns a set of the unbound Parameters in the expression."""
        return self._parameter_symbols.keys()

    @property
    def _names(self) -> dict:
        """Returns a mapping of parameter names to Parameters in the expression."""
        if self._name_map is None:
            self._name_map = {p.name: p for p in self._parameter_symbols}
        return self._name_map

    def conjugate(self) -> "ParameterExpression":
        """Return the conjugate."""
        if self._standalone_param:
            new_op = _INSTRUCTION(_OPCode.CONJ, self)
        else:
            new_op = _INSTRUCTION(_OPCode.CONJ, None)
        new_replay = self._qpy_replay.copy()
        new_replay.append(new_op)
        conjugated = ParameterExpression(
            self._parameter_symbols, symengine.conjugate(self._symbol_expr), _qpy_replay=new_replay
        )
        return conjugated

    def assign(self, parameter, value: ParameterValueType) -> "ParameterExpression":
        """
        Assign one parameter to a value, which can either be numeric or another parameter
        expression.

        Args:
            parameter (Parameter): A parameter in this expression whose value will be updated.
            value: The new value to bind to.

        Returns:
            A new expression parameterized by any parameters which were not bound by assignment.
        """
        if isinstance(value, ParameterExpression):
            return self.subs({parameter: value})
        return self.bind({parameter: value})

    def bind(
        self, parameter_values: dict, allow_unknown_parameters: bool = False
    ) -> "ParameterExpression":
        """Binds the provided set of parameters to their corresponding values.

        Args:
            parameter_values: Mapping of Parameter instances to the numeric value to which
                              they will be bound.
            allow_unknown_parameters: If ``False``, raises an error if ``parameter_values``
                contains Parameters in the keys outside those present in the expression.
                If ``True``, any such parameters are simply ignored.

        Raises:
            CircuitError:
                - If parameter_values contains Parameters outside those in self.
                - If a non-numeric value is passed in parameter_values.
            ZeroDivisionError:
                - If binding the provided values requires division by zero.

        Returns:
            A new expression parameterized by any parameters which were not bound by
            parameter_values.
        """
        if not allow_unknown_parameters:
            self._raise_if_passed_unknown_parameters(parameter_values.keys())
        self._raise_if_passed_nan(parameter_values)

        new_op = _SUBS(parameter_values)
        symbol_values = {}
        for parameter, value in parameter_values.items():
            if (param_expr := self._parameter_symbols.get(parameter)) is not None:
                symbol_values[param_expr] = value

        bound_symbol_expr = self._symbol_expr.subs(symbol_values)

        # Don't use sympy.free_symbols to count remaining parameters here.
        # sympy will in some cases reduce the expression and remove even
        # unbound symbols.
        # e.g. (sympy.Symbol('s') * 0).free_symbols == set()

        free_parameters = self.parameters - parameter_values.keys()
        free_parameter_symbols = {
            p: s for p, s in self._parameter_symbols.items() if p in free_parameters
        }

        if (
            hasattr(bound_symbol_expr, "is_infinite") and bound_symbol_expr.is_infinite
        ) or bound_symbol_expr == float("inf"):
            raise ZeroDivisionError(
                "Binding provided for expression "
                "results in division by zero "
                f"(Expression: {self}, Bindings: {parameter_values})."
            )

        new_replay = self._qpy_replay.copy()
        new_replay.append(new_op)

        return ParameterExpression(
            free_parameter_symbols, bound_symbol_expr, _qpy_replay=new_replay
        )

    def subs(
        self, parameter_map: dict, allow_unknown_parameters: bool = False
    ) -> "ParameterExpression":
        """Returns a new Expression with replacement Parameters.

        Args:
            parameter_map: Mapping from Parameters in self to the ParameterExpression
                           instances with which they should be replaced.
            allow_unknown_parameters: If ``False``, raises an error if ``parameter_map``
                contains Parameters in the keys outside those present in the expression.
                If ``True``, any such parameters are simply ignored.

        Raises:
            CircuitError:
                - If parameter_map contains Parameters outside those in self.
                - If the replacement Parameters in parameter_map would result in
                  a name conflict in the generated expression.

        Returns:
            A new expression with the specified parameters replaced.
        """
        if not allow_unknown_parameters:
            self._raise_if_passed_unknown_parameters(parameter_map.keys())

        inbound_names = {
            p.name: p
            for replacement_expr in parameter_map.values()
            for p in replacement_expr.parameters
        }
        self._raise_if_parameter_names_conflict(inbound_names, parameter_map.keys())
        new_op = _SUBS(parameter_map)

        # Include existing parameters in self not set to be replaced.
        new_parameter_symbols = {
            p: s for p, s in self._parameter_symbols.items() if p not in parameter_map
        }
        symbol_type = symengine.Symbol

        # If new_param is an expr, we'll need to construct a matching sympy expr
        # but with our sympy symbols instead of theirs.
        symbol_map = {}
        for old_param, new_param in parameter_map.items():
            if (old_symbol := self._parameter_symbols.get(old_param)) is not None:
                symbol_map[old_symbol] = new_param._symbol_expr
                for p in new_param.parameters:
                    new_parameter_symbols[p] = symbol_type(p.name)

        substituted_symbol_expr = self._symbol_expr.subs(symbol_map)
        new_replay = self._qpy_replay.copy()
        new_replay.append(new_op)

        return ParameterExpression(
            new_parameter_symbols, substituted_symbol_expr, _qpy_replay=new_replay
        )

    def _raise_if_passed_unknown_parameters(self, parameters):
        unknown_parameters = parameters - self.parameters
        if unknown_parameters:
            raise CircuitError(
                f"Cannot bind Parameters ({[str(p) for p in unknown_parameters]}) not present in "
                "expression."
            )

    def _raise_if_passed_nan(self, parameter_values):
        nan_parameter_values = {
            p: v for p, v in parameter_values.items() if not isinstance(v, numbers.Number)
        }
        if nan_parameter_values:
            raise CircuitError(
                f"Expression cannot bind non-numeric values ({nan_parameter_values})"
            )

    def _raise_if_parameter_names_conflict(self, inbound_parameters, outbound_parameters=None):
        if outbound_parameters is None:
            outbound_parameters = set()
            outbound_names = {}
        else:
            outbound_names = {p.name: p for p in outbound_parameters}

        inbound_names = inbound_parameters
        conflicting_names = []
        for name, param in inbound_names.items():
            if name in self._names and name not in outbound_names:
                if param != self._names[name]:
                    conflicting_names.append(name)
        if conflicting_names:
            raise CircuitError(
                f"Name conflict applying operation for parameters: {conflicting_names}"
            )

    def _apply_operation(
        self,
        operation: Callable,
        other: ParameterValueType,
        reflected: bool = False,
        op_code: _OPCode = None,
    ) -> "ParameterExpression":
        """Base method implementing math operations between Parameters and
        either a constant or a second ParameterExpression.

        Args:
            operation: An operator, such as add, sub, mul, and truediv.
            other: The second argument to be used with self in operation.
            reflected: Optional - The default ordering is "self operator other".
                       If reflected is True, this is switched to "other operator self".
                       For use in e.g. __radd__, ...

        Raises:
            CircuitError:
                - If parameter_map contains Parameters outside those in self.
                - If the replacement Parameters in parameter_map would result in
                  a name conflict in the generated expression.

        Returns:
            A new expression describing the result of the operation.
        """
        self_expr = self._symbol_expr
        if isinstance(other, ParameterExpression):
            self._raise_if_parameter_names_conflict(other._names)
            parameter_symbols = {**self._parameter_symbols, **other._parameter_symbols}
            other_expr = other._symbol_expr
        elif isinstance(other, numbers.Number) and numpy.isfinite(other):
            parameter_symbols = self._parameter_symbols.copy()
            other_expr = other
        else:
            return NotImplemented

        if reflected:
            expr = operation(other_expr, self_expr)
            if op_code in {_OPCode.RSUB, _OPCode.RDIV, _OPCode.RPOW}:
                if self._standalone_param:
                    new_op = _INSTRUCTION(op_code, self, other)
                else:
                    new_op = _INSTRUCTION(op_code, None, other)
            else:
                if self._standalone_param:
                    new_op = _INSTRUCTION(op_code, other, self)
                else:
                    new_op = _INSTRUCTION(op_code, other, None)
        else:
            expr = operation(self_expr, other_expr)
            if self._standalone_param:
                new_op = _INSTRUCTION(op_code, self, other)
            else:
                new_op = _INSTRUCTION(op_code, None, other)
        new_replay = self._qpy_replay.copy()
        new_replay.append(new_op)

        out_expr = ParameterExpression(parameter_symbols, expr, _qpy_replay=new_replay)
        out_expr._name_map = self._names.copy()
        if isinstance(other, ParameterExpression):
            out_expr._names.update(other._names.copy())

        return out_expr

    def gradient(self, param) -> Union["ParameterExpression", complex]:
        """Get the derivative of a parameter expression w.r.t. a specified parameter expression.

        Args:
            param (Parameter): Parameter w.r.t. which we want to take the derivative

        Returns:
            ParameterExpression representing the gradient of param_expr w.r.t. param
            or complex or float number
        """
        # Check if the parameter is contained in the parameter expression
        if param not in self._parameter_symbols.keys():
            # If it is not contained then return 0
            return 0.0

        if self._standalone_param:
            new_op = _INSTRUCTION(_OPCode.GRAD, self, param)
        else:
            new_op = _INSTRUCTION(_OPCode.GRAD, None, param)
        qpy_replay = self._qpy_replay.copy()
        qpy_replay.append(new_op)

        # Compute the gradient of the parameter expression w.r.t. param
        key = self._parameter_symbols[param]
        expr_grad = symengine.Derivative(self._symbol_expr, key)

        # generate the new dictionary of symbols
        # this needs to be done since in the derivative some symbols might disappear (e.g.
        # when deriving linear expression)
        parameter_symbols = {}
        for parameter, symbol in self._parameter_symbols.items():
            if symbol in expr_grad.free_symbols:
                parameter_symbols[parameter] = symbol
        # If the gradient corresponds to a parameter expression then return the new expression.
        if len(parameter_symbols) > 0:
            return ParameterExpression(parameter_symbols, expr=expr_grad, _qpy_replay=qpy_replay)
        # If no free symbols left, return a complex or float gradient
        expr_grad_cplx = complex(expr_grad)
        if expr_grad_cplx.imag != 0:
            return expr_grad_cplx
        else:
            return float(expr_grad)

    def __add__(self, other):
        return self._apply_operation(operator.add, other, op_code=_OPCode.ADD)

    def __radd__(self, other):
        return self._apply_operation(operator.add, other, reflected=True, op_code=_OPCode.ADD)

    def __sub__(self, other):
        return self._apply_operation(operator.sub, other, op_code=_OPCode.SUB)

    def __rsub__(self, other):
        return self._apply_operation(operator.sub, other, reflected=True, op_code=_OPCode.RSUB)

    def __mul__(self, other):
        return self._apply_operation(operator.mul, other, op_code=_OPCode.MUL)

    def __pos__(self):
        return self._apply_operation(operator.mul, 1, op_code=_OPCode.MUL)

    def __neg__(self):
        return self._apply_operation(operator.mul, -1, op_code=_OPCode.MUL)

    def __rmul__(self, other):
        return self._apply_operation(operator.mul, other, reflected=True, op_code=_OPCode.MUL)

    def __truediv__(self, other):
        if other == 0:
            raise ZeroDivisionError("Division of a ParameterExpression by zero.")
        return self._apply_operation(operator.truediv, other, op_code=_OPCode.DIV)

    def __rtruediv__(self, other):
        return self._apply_operation(operator.truediv, other, reflected=True, op_code=_OPCode.RDIV)

    def __pow__(self, other):
        return self._apply_operation(pow, other, op_code=_OPCode.POW)

    def __rpow__(self, other):
        return self._apply_operation(pow, other, reflected=True, op_code=_OPCode.RPOW)

    def _call(self, ufunc, op_code):
        if self._standalone_param:
            new_op = _INSTRUCTION(op_code, self)
        else:
            new_op = _INSTRUCTION(op_code, None)
        new_replay = self._qpy_replay.copy()
        new_replay.append(new_op)
        return ParameterExpression(
            self._parameter_symbols, ufunc(self._symbol_expr), _qpy_replay=new_replay
        )

    def sin(self):
        """Sine of a ParameterExpression"""
        return self._call(symengine.sin, op_code=_OPCode.SIN)

    def cos(self):
        """Cosine of a ParameterExpression"""
        return self._call(symengine.cos, op_code=_OPCode.COS)

    def tan(self):
        """Tangent of a ParameterExpression"""
        return self._call(symengine.tan, op_code=_OPCode.TAN)

    def arcsin(self):
        """Arcsin of a ParameterExpression"""
        return self._call(symengine.asin, op_code=_OPCode.ASIN)

    def arccos(self):
        """Arccos of a ParameterExpression"""
        return self._call(symengine.acos, op_code=_OPCode.ACOS)

    def arctan(self):
        """Arctan of a ParameterExpression"""
        return self._call(symengine.atan, op_code=_OPCode.ATAN)

    def exp(self):
        """Exponential of a ParameterExpression"""
        return self._call(symengine.exp, op_code=_OPCode.EXP)

    def log(self):
        """Logarithm of a ParameterExpression"""
        return self._call(symengine.log, op_code=_OPCode.LOG)

    def sign(self):
        """Sign of a ParameterExpression"""
        return self._call(symengine.sign, op_code=_OPCode.SIGN)

    def __repr__(self):
        return f"{self.__class__.__name__}({str(self)})"

    def __str__(self):
        from sympy import sympify, sstr

        return sstr(sympify(self._symbol_expr), full_prec=False)

    def __complex__(self):
        try:
            return complex(self._symbol_expr)
        # TypeError is for sympy, RuntimeError for symengine
        except (TypeError, RuntimeError) as exc:
            if self.parameters:
                raise TypeError(
                    f"ParameterExpression with unbound parameters ({self.parameters}) "
                    "cannot be cast to a complex."
                ) from None
            raise TypeError("could not cast expression to complex") from exc

    def __float__(self):
        try:
            return float(self._symbol_expr)
        # TypeError is for sympy, RuntimeError for symengine
        except (TypeError, RuntimeError) as exc:
            if self.parameters:
                raise TypeError(
                    f"ParameterExpression with unbound parameters ({self.parameters}) "
                    "cannot be cast to a float."
                ) from None
            # In symengine, if an expression was complex at any time, its type is likely to have
            # stayed "complex" even when the imaginary part symbolically (i.e. exactly)
            # cancelled out.  Sympy tends to more aggressively recognize these as symbolically
            # real.  This second attempt at a cast is a way of unifying the behavior to the
            # more expected form for our users.
            cval = complex(self)
            if cval.imag == 0.0:
                return cval.real
            raise TypeError("could not cast expression to float") from exc

    def __int__(self):
        try:
            return int(self._symbol_expr)
        # TypeError is for backwards compatibility, RuntimeError is raised by symengine
        except RuntimeError as exc:
            if self.parameters:
                raise TypeError(
                    f"ParameterExpression with unbound parameters ({self.parameters}) "
                    "cannot be cast to an int."
                ) from None
            raise TypeError("could not cast expression to int") from exc

    def __hash__(self):
        if not self._parameter_symbols:
            # For fully bound expressions, fall back to the underlying value
            return hash(self.numeric())
        return hash((self._parameter_keys, self._symbol_expr))

    def __copy__(self):
        return self

    def __deepcopy__(self, memo=None):
        return self

    def __abs__(self):
        """Absolute of a ParameterExpression"""
        return self._call(symengine.Abs, _OPCode.ABS)

    def abs(self):
        """Absolute of a ParameterExpression"""
        return self.__abs__()

    def __eq__(self, other):
        """Check if this parameter expression is equal to another parameter expression
           or a fixed value (only if this is a bound expression).
        Args:
            other (ParameterExpression or a number):
                Parameter expression or numeric constant used for comparison
        Returns:
            bool: result of the comparison
        """
        if isinstance(other, ParameterExpression):
            if self.parameters != other.parameters:
                return False
            from sympy import sympify

            return sympify(self._symbol_expr).equals(sympify(other._symbol_expr))
        elif isinstance(other, numbers.Number):
            return len(self.parameters) == 0 and complex(self._symbol_expr) == other
        return False

    def is_real(self):
        """Return whether the expression is real"""
        if not self._symbol_expr.is_real and self._symbol_expr.is_real is not None:
            # Symengine returns false for is_real on the expression if
            # there is a imaginary component (even if that component is 0),
            # but the parameter will evaluate as real. Check that if the
            # expression's is_real attribute returns false that we have a
            # non-zero imaginary
            if self._symbol_expr.imag == 0.0:
                return True
            return False
        return self._symbol_expr.is_real

    def numeric(self) -> int | float | complex:
        """Return a Python number representing this object, using the most restrictive of
        :class:`int`, :class:`float` and :class:`complex` that is valid for this object.

        In general, an :class:`int` is only returned if the expression only involved symbolic
        integers.  If floating-point values were used during the evaluation, the return value will
        be a :class:`float` regardless of whether the represented value is an integer.  This is
        because floating-point values "infect" symbolic computations by their inexact nature, and
        symbolic libraries will use inexact floating-point semantics not exact real-number semantics
        when they are involved.  If you want to assert that all floating-point calculations *were*
        carried out at infinite precision (i.e. :class:`float` could represent every intermediate
        value exactly), you can use :meth:`float.is_integer` to check if the return float represents
        an integer and cast it using :class:`int` if so.  This would be an unusual pattern;
        typically one requires this by only ever using explicitly :class:`~numbers.Rational` objects
        while working with symbolic expressions.

        This is more reliable and performant than using :meth:`is_real` followed by calling
        :class:`float` or :class:`complex`, as in some cases :meth:`is_real` needs to force a
        floating-point evaluation to determine an accurate result to work around bugs in the
        upstream symbolic libraries.

        Returns:
            A Python number representing the object.

        Raises:
            TypeError: if there are unbound parameters.
        """
        if self._parameter_symbols:
            raise TypeError(
                f"Expression with unbound parameters '{self.parameters}' is not numeric"
            )
        if self._symbol_expr.is_integer:
            # Integer evaluation is reliable, as far as we know.
            return int(self._symbol_expr)
        # We've come across several ways in which symengine's general-purpose evaluators
        # introduce spurious imaginary components can get involved in the output.  The most
        # reliable strategy "try it and see" while forcing real floating-point evaluation.
        try:
            real_expr = self._symbol_expr.evalf(real=True)
        except RuntimeError:
            # Symengine returns `complex` if any imaginary floating-point enters at all, even if
            # the result is zero.  The best we can do is detect that and decay to a float.
            out = complex(self._symbol_expr)
            return out.real if out.imag == 0.0 else out
        return float(real_expr)

    def sympify(self):
        """Return symbolic expression as a raw Sympy or Symengine object.

        Symengine is used preferentially; if both are available, the result will always be a
        ``symengine`` object.  Symengine is a separate library but has integration with Sympy.

        .. note::

            This is for interoperability only.  Qiskit will not accept or work with raw Sympy or
            Symegine expressions in its parameters, because they do not contain the tracking
            information used in circuit-parameter binding and assignment.
        """
        return self._symbol_expr


# Redefine the type so external imports get an evaluated reference; Sphinx needs this to understand
# the type hints.
ParameterValueType = Union[ParameterExpression, float]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test the CSPLayout pass"""

import unittest
from time import process_time

from qiskit import QuantumRegister, QuantumCircuit
from qiskit.providers.fake_provider import GenericBackendV2
from qiskit.transpiler import CouplingMap
from qiskit.transpiler.passes import CSPLayout
from qiskit.converters import circuit_to_dag
from qiskit.utils import optionals
from test import QiskitTestCase  # pylint: disable=wrong-import-order

from ..legacy_cmaps import TENERIFE_CMAP, RUESCHLIKON_CMAP, TOKYO_CMAP, YORKTOWN_CMAP


@unittest.skipUnless(optionals.HAS_CONSTRAINT, "needs python-constraint")
class TestCSPLayout(QiskitTestCase):
    """Tests the CSPLayout pass"""

    seed = 42

    def test_2q_circuit_2q_coupling(self):
        """A simple example, without considering the direction
          0 - 1
        qr1 - qr0
        """
        qr = QuantumRegister(2, "qr")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[1], qr[0])  # qr1 -> qr0

        dag = circuit_to_dag(circuit)
        pass_ = CSPLayout(CouplingMap([[0, 1]]), strict_direction=False, seed=self.seed)
        pass_.run(dag)
        layout = pass_.property_set["layout"]

        self.assertEqual(layout[qr[0]], 1)
        self.assertEqual(layout[qr[1]], 0)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "solution found")

    def test_3q_circuit_5q_coupling(self):
        """3 qubits in Tenerife, without considering the direction
            qr1
           /  |
        qr0 - qr2 - 3
              |   /
               4
        """
        cmap5 = TENERIFE_CMAP
        qr = QuantumRegister(3, "qr")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[1], qr[0])  # qr1 -> qr0
        circuit.cx(qr[0], qr[2])  # qr0 -> qr2
        circuit.cx(qr[1], qr[2])  # qr1 -> qr2

        dag = circuit_to_dag(circuit)
        pass_ = CSPLayout(CouplingMap(cmap5), strict_direction=False, seed=self.seed)
        pass_.run(dag)
        layout = pass_.property_set["layout"]

        self.assertEqual(layout[qr[0]], 3)
        self.assertEqual(layout[qr[1]], 2)
        self.assertEqual(layout[qr[2]], 4)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "solution found")

    def test_3q_circuit_5q_coupling_with_target(self):
        """3 qubits in Yorktown, without considering the direction
            qr1
           /  |
        qr0 - qr2 - 3
              |   /
               4
        """
        target = GenericBackendV2(
            num_qubits=5,
            coupling_map=YORKTOWN_CMAP,
        ).target

        qr = QuantumRegister(3, "qr")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[1], qr[0])  # qr1 -> qr0
        circuit.cx(qr[0], qr[2])  # qr0 -> qr2
        circuit.cx(qr[1], qr[2])  # qr1 -> qr2s

        dag = circuit_to_dag(circuit)
        pass_ = CSPLayout(target, strict_direction=False, seed=self.seed)
        pass_.run(dag)
        layout = pass_.property_set["layout"]

        self.assertEqual(layout[qr[0]], 3)
        self.assertEqual(layout[qr[1]], 2)
        self.assertEqual(layout[qr[2]], 4)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "solution found")

    def test_9q_circuit_16q_coupling(self):
        """9 qubits in Rueschlikon, without considering the direction
        q0[1] - q0[0] - q1[3] - q0[3] - q1[0] - q1[1] - q1[2] - 8
          |       |       |       |       |       |       |     |
        q0[2] - q1[4] -- 14 ---- 13 ---- 12 ---- 11 ---- 10 --- 9
        """
        cmap16 = RUESCHLIKON_CMAP

        qr0 = QuantumRegister(4, "q0")
        qr1 = QuantumRegister(5, "q1")
        circuit = QuantumCircuit(qr0, qr1)
        circuit.cx(qr0[1], qr0[2])  # q0[1] -> q0[2]
        circuit.cx(qr0[0], qr1[3])  # q0[0] -> q1[3]
        circuit.cx(qr1[4], qr0[2])  # q1[4] -> q0[2]

        dag = circuit_to_dag(circuit)
        pass_ = CSPLayout(CouplingMap(cmap16), strict_direction=False, seed=self.seed)
        pass_.run(dag)
        layout = pass_.property_set["layout"]

        self.assertEqual(layout[qr0[0]], 9)
        self.assertEqual(layout[qr0[1]], 6)
        self.assertEqual(layout[qr0[2]], 7)
        self.assertEqual(layout[qr0[3]], 5)
        self.assertEqual(layout[qr1[0]], 14)
        self.assertEqual(layout[qr1[1]], 12)
        self.assertEqual(layout[qr1[2]], 1)
        self.assertEqual(layout[qr1[3]], 8)
        self.assertEqual(layout[qr1[4]], 10)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "solution found")

    def test_2q_circuit_2q_coupling_sd(self):
        """A simple example, considering the direction
         0  -> 1
        qr1 -> qr0
        """
        qr = QuantumRegister(2, "qr")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[1], qr[0])  # qr1 -> qr0

        dag = circuit_to_dag(circuit)
        pass_ = CSPLayout(CouplingMap([[0, 1]]), strict_direction=True, seed=self.seed)
        pass_.run(dag)
        layout = pass_.property_set["layout"]

        self.assertEqual(layout[qr[0]], 1)
        self.assertEqual(layout[qr[1]], 0)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "solution found")

    def test_3q_circuit_5q_coupling_sd(self):
        """3 qubits in Tenerife, considering the direction
              qr0
            ↙  ↑
        qr2 ← qr1 ← 3
               ↑  ↙
               4
        """
        cmap5 = TENERIFE_CMAP

        qr = QuantumRegister(3, "qr")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[1], qr[0])  # qr1 -> qr0
        circuit.cx(qr[0], qr[2])  # qr0 -> qr2
        circuit.cx(qr[1], qr[2])  # qr1 -> qr2

        dag = circuit_to_dag(circuit)
        pass_ = CSPLayout(CouplingMap(cmap5), strict_direction=True, seed=self.seed)
        pass_.run(dag)
        layout = pass_.property_set["layout"]

        self.assertEqual(layout[qr[0]], 1)
        self.assertEqual(layout[qr[1]], 2)
        self.assertEqual(layout[qr[2]], 0)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "solution found")

    def test_9q_circuit_16q_coupling_sd(self):
        """9 qubits in Rueschlikon, considering the direction
        q0[1] → q0[0] → q1[3] → q0[3] ← q1[0] ← q1[1] → q1[2] ← 8
          ↓       ↑      ↓      ↓       ↑       ↓        ↓      ↑
        q0[2] ← q1[4] → 14  ←  13   ←  12   →  11   →   10   ←  9
        """
        cmap16 = RUESCHLIKON_CMAP

        qr0 = QuantumRegister(4, "q0")
        qr1 = QuantumRegister(5, "q1")
        circuit = QuantumCircuit(qr0, qr1)
        circuit.cx(qr0[1], qr0[2])  # q0[1] -> q0[2]
        circuit.cx(qr0[0], qr1[3])  # q0[0] -> q1[3]
        circuit.cx(qr1[4], qr0[2])  # q1[4] -> q0[2]

        dag = circuit_to_dag(circuit)
        pass_ = CSPLayout(CouplingMap(cmap16), strict_direction=True, seed=self.seed)
        pass_.run(dag)
        layout = pass_.property_set["layout"]

        self.assertEqual(layout[qr0[0]], 9)
        self.assertEqual(layout[qr0[1]], 6)
        self.assertEqual(layout[qr0[2]], 7)
        self.assertEqual(layout[qr0[3]], 5)
        self.assertEqual(layout[qr1[0]], 14)
        self.assertEqual(layout[qr1[1]], 12)
        self.assertEqual(layout[qr1[2]], 1)
        self.assertEqual(layout[qr1[3]], 10)
        self.assertEqual(layout[qr1[4]], 8)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "solution found")

    def test_5q_circuit_16q_coupling_no_solution(self):
        """5 qubits in Rueschlikon, no solution

        q0[1] ↖     ↗ q0[2]
               q0[0]
        q0[3] ↙     ↘ q0[4]
        """
        cmap16 = RUESCHLIKON_CMAP

        qr = QuantumRegister(5, "q")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[0], qr[1])
        circuit.cx(qr[0], qr[2])
        circuit.cx(qr[0], qr[3])
        circuit.cx(qr[0], qr[4])
        dag = circuit_to_dag(circuit)
        pass_ = CSPLayout(CouplingMap(cmap16), seed=self.seed)
        pass_.run(dag)
        layout = pass_.property_set["layout"]
        self.assertIsNone(layout)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "nonexistent solution")

    @staticmethod
    def create_hard_dag():
        """Creates a particularly hard circuit (returns its dag) for Tokyo"""
        circuit = QuantumCircuit(20)
        circuit.cx(13, 12)
        circuit.cx(6, 0)
        circuit.cx(5, 10)
        circuit.cx(10, 7)
        circuit.cx(5, 12)
        circuit.cx(2, 15)
        circuit.cx(16, 18)
        circuit.cx(6, 4)
        circuit.cx(10, 3)
        circuit.cx(11, 10)
        circuit.cx(18, 16)
        circuit.cx(5, 12)
        circuit.cx(4, 0)
        circuit.cx(18, 16)
        circuit.cx(2, 15)
        circuit.cx(7, 8)
        circuit.cx(9, 6)
        circuit.cx(16, 17)
        circuit.cx(9, 3)
        circuit.cx(14, 12)
        circuit.cx(2, 15)
        circuit.cx(1, 16)
        circuit.cx(5, 3)
        circuit.cx(8, 12)
        circuit.cx(2, 1)
        circuit.cx(5, 3)
        circuit.cx(13, 5)
        circuit.cx(12, 14)
        circuit.cx(12, 13)
        circuit.cx(6, 4)
        circuit.cx(15, 18)
        circuit.cx(15, 18)
        return circuit_to_dag(circuit)

    def test_time_limit(self):
        """Hard to solve situations hit the time limit"""
        dag = TestCSPLayout.create_hard_dag()
        coupling_map = CouplingMap(TOKYO_CMAP)
        pass_ = CSPLayout(coupling_map, call_limit=None, time_limit=1)

        start = process_time()
        pass_.run(dag)
        runtime = process_time() - start

        self.assertLess(runtime, 3)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "time limit reached")

    def test_call_limit(self):
        """Hard to solve situations hit the call limit"""
        dag = TestCSPLayout.create_hard_dag()
        coupling_map = CouplingMap(TOKYO_CMAP)
        pass_ = CSPLayout(coupling_map, call_limit=1, time_limit=None)

        start = process_time()
        pass_.run(dag)
        runtime = process_time() - start

        self.assertLess(runtime, 1)
        self.assertEqual(pass_.property_set["CSPLayout_stop_reason"], "call limit reached")

    def test_seed(self):
        """Different seeds yield different results"""
        seed_1 = 42
        seed_2 = 43

        cmap5 = TENERIFE_CMAP

        qr = QuantumRegister(3, "qr")
        circuit = QuantumCircuit(qr)
        circuit.cx(qr[1], qr[0])  # qr1 -> qr0
        circuit.cx(qr[0], qr[2])  # qr0 -> qr2
        circuit.cx(qr[1], qr[2])  # qr1 -> qr2
        dag = circuit_to_dag(circuit)

        pass_1 = CSPLayout(CouplingMap(cmap5), seed=seed_1)
        pass_1.run(dag)
        layout_1 = pass_1.property_set["layout"]

        pass_2 = CSPLayout(CouplingMap(cmap5), seed=seed_2)
        pass_2.run(dag)
        layout_2 = pass_2.property_set["layout"]

        self.assertNotEqual(layout_1, layout_2)


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Qobj utilities and enums."""

from enum import Enum, IntEnum

from qiskit.utils import deprecate_func


@deprecate_func(
    since="1.2",
    removal_timeline="in the 2.0 release",
    additional_msg="The `Qobj` class and related functionality are part of the deprecated `BackendV1` "
    "workflow,  and no longer necessary for `BackendV2`. If a user workflow requires `Qobj` it likely "
    "relies on deprecated functionality and should be updated to use `BackendV2`.",
)
class QobjType(str, Enum):
    """Qobj.type allowed values."""

    QASM = "QASM"
    PULSE = "PULSE"


class MeasReturnType(str, Enum):
    """PulseQobjConfig meas_return allowed values."""

    AVERAGE = "avg"
    SINGLE = "single"


class MeasLevel(IntEnum):
    """MeasLevel allowed values."""

    RAW = 0
    KERNELED = 1
    CLASSIFIED = 2



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""BasicProvider Backends Test."""

from qiskit.providers.basic_provider.basic_provider import BasicProvider
from qiskit.providers.exceptions import QiskitBackendNotFoundError
from test import QiskitTestCase  # pylint: disable=wrong-import-order


class TestBasicProviderBackends(QiskitTestCase):
    """Qiskit BasicProvider Backends (Object) Tests."""

    def setUp(self):
        super().setUp()
        with self.assertWarns(DeprecationWarning):
            self.provider = BasicProvider()
        self.backend_name = "basic_simulator"

    def test_backends(self):
        """Test the provider has backends."""
        backends = self.provider.backends()
        self.assertTrue(len(backends) > 0)

    def test_get_backend(self):
        """Test getting a backend from the provider."""
        with self.assertWarns(DeprecationWarning):
            backend = self.provider.get_backend(name=self.backend_name)
        self.assertEqual(backend.name, self.backend_name)

    def test_aliases_fail(self):
        """Test a failing backend lookup."""
        with self.assertWarns(DeprecationWarning):
            self.assertRaises(QiskitBackendNotFoundError, BasicProvider().get_backend, "bad_name")



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Cancel back-to-back ``cx`` gates in dag."""

from qiskit.transpiler.basepasses import TransformationPass
from qiskit.transpiler.passes.utils import control_flow
from qiskit.utils.deprecation import deprecate_func


class CXCancellation(TransformationPass):
    """Cancel back-to-back ``cx`` gates in dag."""

    @deprecate_func(
        additional_msg=(
            "Instead, use :class:`~.InverseCancellation`, which is a more generic pass."
        ),
        since="1.1.0",
    )
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @control_flow.trivial_recurse
    def run(self, dag):
        """Run the CXCancellation pass on `dag`.

        Args:
            dag (DAGCircuit): the directed acyclic graph to run on.

        Returns:
            DAGCircuit: Transformed DAG.
        """
        cx_runs = dag.collect_runs(["cx"])
        for cx_run in cx_runs:
            # Although one `cx_run` will always have the same set of qubits,
            # the order of these qubits matters (CXi_j != CXj_i).
            # Therefore, we have to partition each run into chunks where the qargs
            # are in the same order, and only cancel out CXs within each chunk.
            partitions = []
            chunk = []
            for i in range(len(cx_run) - 1):
                chunk.append(cx_run[i])
                if cx_run[i].qargs != cx_run[i + 1].qargs:
                    partitions.append(chunk)
                    chunk = []
            chunk.append(cx_run[-1])
            partitions.append(chunk)

            # Remove an even number of CX gates from each chunk
            for chunk in partitions:
                if len(chunk) % 2 == 0:
                    dag.remove_op_node(chunk[0])
                for node in chunk[1:]:
                    dag.remove_op_node(node)

        return dag



# This code is part of Qiskit.
#
# (C) Copyright IBM 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""
A 27 qubit fake :class:`.BackendV1` with pulse capabilities.
"""

from .fake_27q_pulse_v1 import Fake27QPulseV1



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test cases for Pulse Macro functions."""

from qiskit.pulse import (
    Schedule,
    AcquireChannel,
    Acquire,
    InstructionScheduleMap,
    MeasureChannel,
    MemorySlot,
    GaussianSquare,
    Play,
)
from qiskit.pulse import macros
from qiskit.pulse.exceptions import PulseError
from qiskit.providers.fake_provider import FakeOpenPulse2Q, Fake27QPulseV1, GenericBackendV2
from test import QiskitTestCase  # pylint: disable=wrong-import-order
from qiskit.utils.deprecate_pulse import decorate_test_methods, ignore_pulse_deprecation_warnings


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestMeasure(QiskitTestCase):
    """Pulse measure macro."""

    @ignore_pulse_deprecation_warnings
    def setUp(self):
        super().setUp()
        with self.assertWarns(DeprecationWarning):
            self.backend = FakeOpenPulse2Q()
            self.backend_v1 = Fake27QPulseV1()

        self.inst_map = self.backend.defaults().instruction_schedule_map
        with self.assertWarns(DeprecationWarning):
            self.backend_v2 = GenericBackendV2(
                num_qubits=27,
                calibrate_instructions=self.backend_v1.defaults().instruction_schedule_map,
                seed=42,
            )

    def test_measure(self):
        """Test macro - measure."""
        sched = macros.measure(qubits=[0], backend=self.backend)
        expected = Schedule(
            self.inst_map.get("measure", [0, 1]).filter(channels=[MeasureChannel(0)]),
            Acquire(10, AcquireChannel(0), MemorySlot(0)),
        )
        self.assertEqual(sched.instructions, expected.instructions)

    def test_measure_sched_with_qubit_mem_slots(self):
        """Test measure with custom qubit_mem_slots."""
        sched = macros.measure(qubits=[0], backend=self.backend, qubit_mem_slots={0: 1})
        expected = Schedule(
            self.inst_map.get("measure", [0, 1]).filter(channels=[MeasureChannel(0)]),
            Acquire(10, AcquireChannel(0), MemorySlot(1)),
        )
        self.assertEqual(sched.instructions, expected.instructions)

    def test_measure_sched_with_meas_map(self):
        """Test measure with custom meas_map as list and dict."""
        sched_with_meas_map_list = macros.measure(
            qubits=[0], backend=self.backend, meas_map=[[0, 1]]
        )
        sched_with_meas_map_dict = macros.measure(
            qubits=[0], backend=self.backend, meas_map={0: [0, 1], 1: [0, 1]}
        )
        expected = Schedule(
            self.inst_map.get("measure", [0, 1]).filter(channels=[MeasureChannel(0)]),
            Acquire(10, AcquireChannel(0), MemorySlot(0)),
        )
        self.assertEqual(sched_with_meas_map_list.instructions, expected.instructions)
        self.assertEqual(sched_with_meas_map_dict.instructions, expected.instructions)

    def test_measure_with_custom_inst_map(self):
        """Test measure with custom inst_map, meas_map with measure_name."""
        q0_sched = Play(GaussianSquare(1200, 1, 0.4, 1150), MeasureChannel(0))
        q0_sched += Acquire(1200, AcquireChannel(0), MemorySlot(0))
        inst_map = InstructionScheduleMap()
        inst_map.add("my_sched", 0, q0_sched)
        sched = macros.measure(
            qubits=[0], measure_name="my_sched", inst_map=inst_map, meas_map=[[0]]
        )
        self.assertEqual(sched.instructions, q0_sched.instructions)

        with self.assertRaises(PulseError):
            macros.measure(qubits=[0], measure_name="name", inst_map=inst_map, meas_map=[[0]])

    def test_fail_measure(self):
        """Test failing measure."""
        with self.assertRaises(PulseError):
            macros.measure(qubits=[0], meas_map=self.backend.configuration().meas_map)
        with self.assertRaises(PulseError):
            macros.measure(qubits=[0], inst_map=self.inst_map)

    def test_measure_v2(self):
        """Test macro - measure with backendV2."""
        sched = macros.measure(qubits=[0], backend=self.backend_v2)
        with self.assertWarns(DeprecationWarning):
            expected = self.backend_v2.target.get_calibration("measure", (0,)).filter(
                channels=[MeasureChannel(0), AcquireChannel(0)]
            )
        self.assertEqual(sched.instructions, expected.instructions)

    def test_measure_v2_sched_with_qubit_mem_slots(self):
        """Test measure with backendV2 and custom qubit_mem_slots."""
        sched = macros.measure(qubits=[0], backend=self.backend_v2, qubit_mem_slots={0: 2})
        with self.assertWarns(DeprecationWarning):
            expected = self.backend_v2.target.get_calibration("measure", (0,)).filter(
                channels=[
                    MeasureChannel(0),
                ]
            )
        measure_duration = expected.filter(instruction_types=[Play]).duration
        expected += Acquire(measure_duration, AcquireChannel(0), MemorySlot(2))
        self.assertEqual(sched.instructions, expected.instructions)

    def test_measure_v2_sched_with_meas_map(self):
        """Test measure with backendV2 custom meas_map as list and dict."""
        sched_with_meas_map_list = macros.measure(
            qubits=[0], backend=self.backend_v2, meas_map=[[0, 1]]
        )
        sched_with_meas_map_dict = macros.measure(
            qubits=[0], backend=self.backend_v2, meas_map={0: [0, 1], 1: [0, 1]}
        )
        with self.assertWarns(DeprecationWarning):
            expected = self.backend_v2.target.get_calibration("measure", (0,)).filter(
                channels=[
                    MeasureChannel(0),
                ]
            )
        measure_duration = expected.filter(instruction_types=[Play]).duration
        expected += Acquire(measure_duration, AcquireChannel(0), MemorySlot(0))
        self.assertEqual(sched_with_meas_map_list.instructions, expected.instructions)
        self.assertEqual(sched_with_meas_map_dict.instructions, expected.instructions)

    def test_multiple_measure_v2(self):
        """Test macro - multiple qubit measure with backendV2."""
        sched = macros.measure(qubits=[0, 1], backend=self.backend_v2)
        with self.assertWarns(DeprecationWarning):
            expected = self.backend_v2.target.get_calibration("measure", (0,)).filter(
                channels=[
                    MeasureChannel(0),
                ]
            )
            expected += self.backend_v2.target.get_calibration("measure", (1,)).filter(
                channels=[
                    MeasureChannel(1),
                ]
            )
        measure_duration = expected.filter(instruction_types=[Play]).duration
        expected += Acquire(measure_duration, AcquireChannel(0), MemorySlot(0))
        expected += Acquire(measure_duration, AcquireChannel(1), MemorySlot(1))
        self.assertEqual(sched.instructions, expected.instructions)

    def test_output_with_measure_v1_and_measure_v2(self):
        """Test make outputs of measure_v1 and measure_v2 consistent."""
        sched_measure_v1 = macros.measure(qubits=[0, 1], backend=self.backend_v1)
        sched_measure_v2 = macros.measure(qubits=[0, 1], backend=self.backend_v2)

        self.assertEqual(sched_measure_v1.instructions, sched_measure_v2.instructions)

    def test_output_with_measure_v1_and_measure_v2_sched_with_qubit_mem_slots(self):
        """Test make outputs of measure_v1 and measure_v2 with custom qubit_mem_slots consistent."""
        sched_measure_v1 = macros.measure(
            qubits=[0], backend=self.backend_v1, qubit_mem_slots={0: 2}
        )
        sched_measure_v2 = macros.measure(
            qubits=[0], backend=self.backend_v2, qubit_mem_slots={0: 2}
        )
        self.assertEqual(sched_measure_v1.instructions, sched_measure_v2.instructions)

    def test_output_with_measure_v1_and_measure_v2_sched_with_meas_map(self):
        """Test make outputs of measure_v1 and measure_v2
        with custom meas_map as list and dict consistent."""
        with self.assertWarns(DeprecationWarning):
            backend = Fake27QPulseV1()
        num_qubits_list_measure_v1 = list(range(backend.configuration().num_qubits))
        num_qubits_list_measure_v2 = list(range(self.backend_v2.num_qubits))
        sched_with_meas_map_list_v1 = macros.measure(
            qubits=[0], backend=self.backend_v1, meas_map=[num_qubits_list_measure_v1]
        )
        sched_with_meas_map_dict_v1 = macros.measure(
            qubits=[0],
            backend=self.backend_v1,
            meas_map={0: num_qubits_list_measure_v1, 1: num_qubits_list_measure_v1},
        )
        sched_with_meas_map_list_v2 = macros.measure(
            qubits=[0], backend=self.backend_v2, meas_map=[num_qubits_list_measure_v2]
        )
        sched_with_meas_map_dict_v2 = macros.measure(
            qubits=[0],
            backend=self.backend_v2,
            meas_map={0: num_qubits_list_measure_v2, 1: num_qubits_list_measure_v2},
        )
        self.assertEqual(
            sched_with_meas_map_list_v1.instructions,
            sched_with_meas_map_list_v2.instructions,
        )
        self.assertEqual(
            sched_with_meas_map_dict_v1.instructions,
            sched_with_meas_map_dict_v2.instructions,
        )

    def test_output_with_multiple_measure_v1_and_measure_v2(self):
        """Test macro - consistent output of multiple qubit measure with backendV1 and backendV2."""
        sched_measure_v1 = macros.measure(qubits=[0, 1], backend=self.backend_v1)
        sched_measure_v2 = macros.measure(qubits=[0, 1], backend=self.backend_v2)
        self.assertEqual(sched_measure_v1.instructions, sched_measure_v2.instructions)


@decorate_test_methods(ignore_pulse_deprecation_warnings)
class TestMeasureAll(QiskitTestCase):
    """Pulse measure all macro."""

    @ignore_pulse_deprecation_warnings
    def setUp(self):
        super().setUp()
        with self.assertWarns(DeprecationWarning):
            self.backend_v1 = FakeOpenPulse2Q()
        self.inst_map = self.backend_v1.defaults().instruction_schedule_map
        with self.assertWarns(DeprecationWarning):
            self.backend_v2 = GenericBackendV2(
                num_qubits=2,
                calibrate_instructions=self.backend_v1.defaults().instruction_schedule_map,
                seed=42,
            )

    def test_measure_all(self):
        """Test measure_all function."""
        sched = macros.measure_all(self.backend_v1)
        expected = Schedule(self.inst_map.get("measure", [0, 1]))
        self.assertEqual(sched.instructions, expected.instructions)

    def test_measure_all_v2(self):
        """Test measure_all function with backendV2."""
        sched = macros.measure_all(self.backend_v1)
        expected = Schedule(
            self.inst_map.get("measure", list(range(self.backend_v1.configuration().num_qubits)))
        )
        self.assertEqual(sched.instructions, expected.instructions)

    def test_output_of_measure_all_with_backend_v1_and_v2(self):
        """Test make outputs of measure_all with backendV1 and backendV2 consistent."""
        sched_measure_v1 = macros.measure_all(backend=self.backend_v1)
        sched_measure_v2 = macros.measure_all(backend=self.backend_v2)
        self.assertEqual(sched_measure_v1.instructions, sched_measure_v2.instructions)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Instruction collection.
"""

from __future__ import annotations

from collections.abc import MutableSequence
from typing import Callable

from qiskit.circuit.exceptions import CircuitError
from qiskit.utils import deprecate_func
from .classicalregister import Clbit, ClassicalRegister
from .operation import Operation
from .quantumcircuitdata import CircuitInstruction


class InstructionSet:
    """Instruction collection, and their contexts."""

    __slots__ = ("_instructions", "_requester")

    def __init__(  # pylint: disable=bad-docstring-quotes
        self,
        *,
        resource_requester: Callable[..., ClassicalRegister | Clbit] | None = None,
    ):
        """New collection of instructions.

        The context (``qargs`` and ``cargs`` that each instruction is attached to) is also stored
        separately for each instruction.

        Args:
            resource_requester: A callable that takes in the classical resource used in the
                condition, verifies that it is present in the attached circuit, resolves any indices
                into concrete :obj:`.Clbit` instances, and returns the concrete resource.  If this
                is not given, specifying a condition with an index is forbidden, and all concrete
                :obj:`.Clbit` and :obj:`.ClassicalRegister` resources will be assumed to be valid.

                .. note::

                    The callback ``resource_requester`` is called once for each call to
                    :meth:`.c_if`, and assumes that a call implies that the resource will now be
                    used.  It may throw an error if the resource is not valid for usage.

        """
        self._instructions: list[
            CircuitInstruction | (MutableSequence[CircuitInstruction], int)
        ] = []
        self._requester = resource_requester

    def __len__(self):
        """Return number of instructions in set"""
        return len(self._instructions)

    def __getitem__(self, i):
        """Return instruction at index"""
        inst = self._instructions[i]
        if isinstance(inst, CircuitInstruction):
            return inst
        data, idx = inst
        return data[idx]

    def add(self, instruction, qargs=None, cargs=None):
        """Add an instruction and its context (where it is attached)."""
        if not isinstance(instruction, CircuitInstruction):
            if not isinstance(instruction, Operation):
                raise CircuitError("attempt to add non-Operation to InstructionSet")
            if qargs is None or cargs is None:
                raise CircuitError("missing qargs or cargs in old-style InstructionSet.add")
            instruction = CircuitInstruction(instruction, tuple(qargs), tuple(cargs))
        self._instructions.append(instruction)

    def _add_ref(self, data: MutableSequence[CircuitInstruction], pos: int):
        """Add a reference to an instruction and its context within a mutable sequence.
        Updates to the instruction set will modify the specified sequence in place."""
        self._instructions.append((data, pos))

    def inverse(self, annotated: bool = False):
        """Invert all instructions.

        .. note::
            It is preferable to take the inverse *before* appending the gate(s) to the circuit.
        """
        for i, instruction in enumerate(self._instructions):
            if isinstance(instruction, CircuitInstruction):
                self._instructions[i] = instruction.replace(
                    operation=instruction.operation.inverse(annotated=annotated)
                )
            else:
                data, idx = instruction
                instruction = data[idx]
                data[idx] = instruction.replace(
                    operation=instruction.operation.inverse(annotated=annotated)
                )
        return self

    @deprecate_func(since="1.3.0", removal_timeline="in 2.0.0")
    def c_if(self, classical: Clbit | ClassicalRegister | int, val: int) -> "InstructionSet":
        """Set a classical equality condition on all the instructions in this set between the
        :obj:`.ClassicalRegister` or :obj:`.Clbit` ``classical`` and value ``val``.

        .. note::
            You should prefer to use the :meth:`.QuantumCircuit.if_test` builder interface, rather
            than using this method.

        .. note::

            This is a setter method, not an additive one.  Calling this multiple times will silently
            override any previously set condition on any of the contained instructions; it does not
            stack.

        Args:
            classical: the classical resource the equality condition should be on.  If this is given
                as an integer, it will be resolved into a :obj:`.Clbit` using the same conventions
                as the circuit these instructions are attached to.
            val: the value the classical resource should be equal to.

        Returns:
            This same instance of :obj:`.InstructionSet`, but now mutated to have the given equality
            condition.

        Raises:
            CircuitError: if the passed classical resource is invalid, or otherwise not resolvable
                to a concrete resource that these instructions are permitted to access.
        """
        if self._requester is None and not isinstance(classical, (Clbit, ClassicalRegister)):
            raise CircuitError(
                "Cannot pass an index as a condition variable without specifying a requester"
                " when creating this InstructionSet."
            )
        if self._requester is not None:
            classical = self._requester(classical)
        for idx, instruction in enumerate(self._instructions):
            if isinstance(instruction, CircuitInstruction):
                updated = instruction.operation.c_if(classical, val)
                self._instructions[idx] = instruction.replace(operation=updated)
            else:
                data, idx = instruction
                instruction = data[idx]
                data[idx] = instruction.replace(
                    operation=instruction.operation.c_if(classical, val)
                )
        return self

    # Legacy support for properties.  Added in Terra 0.21 to support the internal switch in
    # `QuantumCircuit.data` from the 3-tuple to `CircuitInstruction`.

    def _instructions_iter(self):
        return (i if isinstance(i, CircuitInstruction) else i[0][i[1]] for i in self._instructions)

    @property
    def instructions(self):
        """Legacy getter for the instruction components of an instruction set.  This does not
        support mutation."""
        return [instruction.operation for instruction in self._instructions_iter()]

    @property
    def qargs(self):
        """Legacy getter for the qargs components of an instruction set.  This does not support
        mutation."""
        return [list(instruction.qubits) for instruction in self._instructions_iter()]

    @property
    def cargs(self):
        """Legacy getter for the cargs components of an instruction set.  This does not support
        mutation."""
        return [list(instruction.clbits) for instruction in self._instructions_iter()]



# This code is part of Qiskit.
#
# (C) Copyright IBM 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.
"""
Circuit synthesis for a QFT circuit.
"""

from __future__ import annotations
import warnings
import numpy as np
from qiskit.circuit.quantumcircuit import QuantumCircuit


def synth_qft_full(
    num_qubits: int,
    do_swaps: bool = True,
    approximation_degree: int = 0,
    insert_barriers: bool = False,
    inverse: bool = False,
    name: str | None = None,
) -> QuantumCircuit:
    """Construct a circuit for the Quantum Fourier Transform using all-to-all connectivity.

    .. note::

        With the default value of ``do_swaps = True``, this synthesis algorithm creates a
        circuit that faithfully implements the QFT operation. This circuit contains a sequence
        of swap gates at the end, corresponding to reversing the order of its output qubits.
        In some applications this reversal permutation can be avoided. Setting ``do_swaps = False``
        creates a circuit without this reversal permutation, at the expense that this circuit
        implements the "QFT-with-reversal" instead of QFT. Alternatively, the
        :class:`~.ElidePermutations` transpiler pass is able to remove these swap gates.

    Args:
        num_qubits: The number of qubits on which the Quantum Fourier Transform acts.
        do_swaps: Whether to synthesize the "QFT" or the "QFT-with-reversal" operation.
        approximation_degree: The degree of approximation (0 for no approximation).
            It is possible to implement the QFT approximately by ignoring
            controlled-phase rotations with the angle beneath a threshold. This is discussed
            in more detail in https://arxiv.org/abs/quant-ph/9601018 or
            https://arxiv.org/abs/quant-ph/0403071.
        insert_barriers: If ``True``, barriers are inserted for improved visualization.
        inverse: If ``True``, the inverse Quantum Fourier Transform is constructed.
        name: The name of the circuit.

    Returns:
        A circuit implementing the QFT operation.

    """
    _warn_if_precision_loss(num_qubits - approximation_degree - 1)
    circuit = QuantumCircuit(num_qubits, name=name)

    for j in reversed(range(num_qubits)):
        circuit.h(j)
        num_entanglements = max(0, j - max(0, approximation_degree - (num_qubits - j - 1)))
        for k in reversed(range(j - num_entanglements, j)):
            # Use negative exponents so that the angle safely underflows to zero, rather than
            # using a temporary variable that overflows to infinity in the worst case.
            lam = np.pi * (2.0 ** (k - j))
            circuit.cp(lam, j, k)

        if insert_barriers:
            circuit.barrier()

    if do_swaps:
        for i in range(num_qubits // 2):
            circuit.swap(i, num_qubits - i - 1)

    if inverse:
        circuit = circuit.inverse()

    return circuit


def _warn_if_precision_loss(max_num_entanglements):
    """Issue a warning if constructing the circuit will lose precision.

    If we need an angle smaller than ``pi * 2**-1022``, we start to lose precision by going into
    the subnormal numbers.  We won't lose _all_ precision until an exponent of about 1075, but
    beyond 1022 we're using fractional bits to represent leading zeros.
    """
    if max_num_entanglements > -np.finfo(float).minexp:  # > 1022 for doubles.
        warnings.warn(
            "precision loss in QFT."
            f" The rotation needed to represent {max_num_entanglements} entanglements"
            " is smaller than the smallest normal floating-point number.",
            category=RuntimeWarning,
            stacklevel=4,
        )



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2020.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Test library of multi-controlled multi-target circuits."""

import unittest
from ddt import ddt, data, unpack
import numpy as np

from qiskit.exceptions import QiskitError
from qiskit.compiler import transpile
from qiskit.circuit import QuantumCircuit, QuantumRegister, Parameter
from qiskit.circuit.library import (
    MCMT,
    MCMTVChain,
    CHGate,
    XGate,
    ZGate,
    RYGate,
    CXGate,
    CZGate,
    MCMTGate,
    GlobalPhaseGate,
    SwapGate,
)
from qiskit.circuit._utils import _compute_control_matrix
from qiskit.converters import circuit_to_dag
from qiskit.quantum_info import Statevector
from qiskit.quantum_info.states import state_fidelity
from qiskit.quantum_info.operators.operator_utils import _equal_with_ancillas
from qiskit.transpiler.passes import HighLevelSynthesis, HLSConfig
from qiskit.synthesis.multi_controlled import synth_mcmt_vchain
from qiskit.quantum_info import Operator
from test import QiskitTestCase  # pylint: disable=wrong-import-order


@ddt
class TestMCMT(QiskitTestCase):
    """Test the multi-controlled multi-target circuit."""

    @data(MCMT, MCMTVChain)
    def test_mcmt_as_normal_control(self, mcmt_class):
        """Test that the MCMT can act as normal control gate."""
        qc = QuantumCircuit(2)
        with self.assertWarns(DeprecationWarning):
            mcmt = mcmt_class(gate=CHGate(), num_ctrl_qubits=1, num_target_qubits=1)
        qc = qc.compose(mcmt, [0, 1])

        ref = QuantumCircuit(2)
        ref.ch(0, 1)

        self.assertEqual(len(qc.data), 1)

        basis = ["u", "cx"]
        self.assertEqual(transpile(qc, basis_gates=basis), transpile(ref, basis_gates=basis))

    def test_missing_qubits(self):
        """Test that an error is raised if qubits are missing."""
        with self.subTest(msg="no control qubits"):
            with self.assertWarns(DeprecationWarning):
                with self.assertRaises(AttributeError):
                    _ = MCMT(XGate(), num_ctrl_qubits=0, num_target_qubits=1)

        with self.subTest(msg="no target qubits"):
            with self.assertWarns(DeprecationWarning):
                with self.assertRaises(AttributeError):
                    _ = MCMT(ZGate(), num_ctrl_qubits=4, num_target_qubits=0)

    def test_different_gate_types(self):
        """Test the different supported input types for the target gate."""
        x_circ = QuantumCircuit(1)
        x_circ.x(0)
        for input_gate in [x_circ, QuantumCircuit.cx, QuantumCircuit.x, "cx", "x", CXGate()]:
            with self.subTest(input_gate=input_gate):
                with self.assertWarns(DeprecationWarning):
                    mcmt = MCMT(input_gate, 2, 2)
                if isinstance(input_gate, QuantumCircuit):
                    self.assertEqual(mcmt.gate.definition[0].operation, XGate())
                    self.assertEqual(len(mcmt.gate.definition), 1)
                else:
                    self.assertEqual(mcmt.gate, XGate())

    def test_mcmt_v_chain_ancilla_test(self):
        """Test too few and too many ancillas for the MCMT V-chain mode."""
        with self.subTest(msg="insufficient number of auxiliary qubits on gate"):
            qc = QuantumCircuit(5)
            with self.assertWarns(DeprecationWarning):
                mcmt = MCMTVChain(ZGate(), 3, 1)
            with self.assertRaises(QiskitError):
                qc.append(mcmt, range(5))

        with self.subTest(msg="too many auxiliary qubits on gate"):
            qc = QuantumCircuit(9)
            with self.assertWarns(DeprecationWarning):
                mcmt = MCMTVChain(ZGate(), 3, 1)
            with self.assertRaises(QiskitError):
                qc.append(mcmt, range(9))

    @data(
        [CZGate(), 1, 1],
        [CHGate(), 1, 1],
        [CZGate(), 3, 3],
        [CHGate(), 3, 3],
        [CZGate(), 1, 5],
        [CHGate(), 1, 5],
        [CZGate(), 5, 1],
        [CHGate(), 5, 1],
    )
    @unpack
    def test_mcmt_v_chain_simulation(self, cgate, num_controls, num_targets):
        """Test the MCMT V-chain implementation test on a simulation."""
        controls = QuantumRegister(num_controls)
        targets = QuantumRegister(num_targets)

        subsets = [tuple(range(i)) for i in range(num_controls + 1)]
        for subset in subsets:
            qc = QuantumCircuit(targets, controls)
            # Initialize all targets to 1, just to be sure that
            # the generic gate has some effect (f.e. Z gate has no effect
            # on a 0 state)
            qc.x(targets)

            num_ancillas = max(0, num_controls - 1)

            if num_ancillas > 0:
                ancillas = QuantumRegister(num_ancillas)
                qc.add_register(ancillas)
                qubits = controls[:] + targets[:] + ancillas[:]
            else:
                qubits = controls[:] + targets[:]

            for i in subset:
                qc.x(controls[i])

            with self.assertWarns(DeprecationWarning):
                mcmt = MCMTVChain(cgate, num_controls, num_targets)
            qc.compose(mcmt, qubits, inplace=True)

            for i in subset:
                qc.x(controls[i])

            vec = Statevector.from_label("0" * qc.num_qubits).evolve(qc)

            # target register is initially |11...1>, with length equal to 2**(n_targets)
            vec_exp = np.array([0] * (2 ** (num_targets) - 1) + [1])

            if isinstance(cgate, CZGate):
                # Z gate flips the last qubit only if it's applied an odd number of times
                if len(subset) == num_controls and (num_controls % 2) == 1:
                    vec_exp[-1] = -1
            elif isinstance(cgate, CHGate):
                # if all the control qubits have been activated,
                # we repeatedly apply the kronecker product of the Hadamard
                # with itself and then multiply the results for the original
                # state of the target qubits
                if len(subset) == num_controls:
                    h_i = 1 / np.sqrt(2) * np.array([[1, 1], [1, -1]])
                    h_tot = np.array([1])
                    for _ in range(num_targets):
                        h_tot = np.kron(h_tot, h_i)
                    vec_exp = np.dot(h_tot, vec_exp)
            else:
                raise ValueError(f"Test not implement for gate: {cgate}")

            # append the remaining part of the state
            vec_exp = np.concatenate(
                (vec_exp, [0] * (2 ** (num_controls + num_ancillas + num_targets) - vec_exp.size))
            )
            f_i = state_fidelity(vec, vec_exp)
            self.assertAlmostEqual(f_i, 1)

    def test_default_plugin(self):
        """Test the default behavior of the plugin."""
        num_controls = 5
        num_target = 2
        num_vchain_ancillas = num_controls - 1

        gate = XGate()
        mcmt = MCMTGate(gate, num_controls, num_target)

        # make sure MCX-synthesis does not use ancilla qubits
        config = HLSConfig(mcx=["noaux_v24"])
        hls = HighLevelSynthesis(hls_config=config)

        # test a decomposition without sufficient ancillas for MCMT V-chain
        with self.subTest(msg="insufficient auxiliaries"):
            circuit = QuantumCircuit(num_controls + num_target + num_vchain_ancillas - 1)
            circuit.append(mcmt, range(mcmt.num_qubits))

            synthesized = hls(circuit)
            num_idle = len(list(circuit_to_dag(synthesized).idle_wires()))

            self.assertEqual(num_idle, num_vchain_ancillas - 1)

        # test with enough auxiliary qubits available
        with self.subTest(msg="enough auxiliaries"):
            circuit = QuantumCircuit(num_controls + num_target + num_vchain_ancillas)
            circuit.append(mcmt, range(mcmt.num_qubits))

            synthesized = hls(circuit)
            num_idle = len(list(circuit_to_dag(synthesized).idle_wires()))

            self.assertEqual(num_idle, 0)

    def test_explicit_plugin(self):
        """Test explicitly setting the plugin."""
        num_controls = 5
        num_target = 2
        num_vchain_ancillas = num_controls - 1

        gate = XGate()
        mcmt = MCMTGate(gate, num_controls, num_target)

        circuit = QuantumCircuit(num_controls + num_target + num_vchain_ancillas)
        circuit.append(mcmt, range(mcmt.num_qubits))

        # test a decomposition without sufficient ancillas for MCMT V-chain
        with self.subTest(msg="force default decomposition"):
            config = HLSConfig(mcmt=["noaux"], mcx=["noaux_v24"])

            synthesized = transpile(circuit, hls_config=config)
            num_idle = len(list(circuit_to_dag(synthesized).idle_wires()))

            self.assertEqual(num_idle, num_vchain_ancillas)

        # test with enough auxiliary qubits available
        with self.subTest(msg="enough auxiliaries"):
            config = HLSConfig(mcmt=["vchain"])

            synthesized = transpile(circuit, hls_config=config)
            num_idle = len(list(circuit_to_dag(synthesized).idle_wires()))

            self.assertEqual(num_idle, 0)

    def test_vchain_open_control(self):
        """Test the V-chain with open controls."""
        num_ctrl_qubits = 2
        ctrl_state = None
        gate = XGate()

        synthesized = synth_mcmt_vchain(
            gate, num_ctrl_qubits, num_target_qubits=1, ctrl_state=ctrl_state
        )
        result = Operator(synthesized)

        # Build the expected matrix by adding padding with the identity operator.
        # This could be potentially moved into the _equal_with_ancillas util.
        action = _compute_control_matrix(gate.to_matrix(), num_ctrl_qubits, ctrl_state)
        pad = np.eye(2 ** (num_ctrl_qubits - 1))
        expected = Operator(np.kron(pad, action))

        self.assertTrue(
            _equal_with_ancillas(
                result,
                expected,
                ancilla_qubits=list(range(num_ctrl_qubits + 1, synthesized.num_qubits)),
                ignore_phase=True,
            )
        )

    def test_invalid_base_gate_width(self):
        """Test only 1-qubit base gates are accepted."""
        for gate in [GlobalPhaseGate(0.2), SwapGate()]:
            with self.subTest(gate=gate):
                with self.assertRaises(ValueError):
                    _ = MCMTGate(gate, 10, 2)

    def test_invalid_base_gate_width_synthfun(self):
        """Test only 1-qubit base gates are accepted."""
        for gate in [GlobalPhaseGate(0.2), SwapGate()]:
            with self.subTest(gate=gate):
                with self.assertRaises(ValueError):
                    _ = synth_mcmt_vchain(gate, 10, 2)

    def test_gate_with_parameters_vchain(self):
        """Test a gate with parameters as base gate."""
        theta = Parameter("th")
        gate = RYGate(theta)
        num_target = 3
        circuit = synth_mcmt_vchain(gate, num_ctrl_qubits=10, num_target_qubits=num_target)

        self.assertEqual(circuit.count_ops().get("cry", 0), num_target)
        self.assertEqual(circuit.num_parameters, 1)
        self.assertIs(circuit.parameters[0], theta)

    def test_mcmt_circuit_as_gate(self):
        """Test the MCMT plugin is only triggered for the gate, not the same-named circuit.

        Regression test of #13563.
        """
        circuit = QuantumCircuit(2)
        gate = RYGate(0.1)
        with self.assertWarns(DeprecationWarning):
            mcmt = MCMT(gate=gate, num_ctrl_qubits=1, num_target_qubits=1)
        circuit.append(mcmt, circuit.qubits)  # append the MCMT circuit as gate called "MCMT"

        transpiled = transpile(circuit, basis_gates=["u", "cx"])
        self.assertEqual(Operator(transpiled), Operator(gate.control(1)))


if __name__ == "__main__":
    unittest.main()



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2018.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
===================================================
Transpiler Passes (:mod:`qiskit.transpiler.passes`)
===================================================

.. currentmodule:: qiskit.transpiler.passes

Layout Selection (Placement)
============================

.. autosummary::
   :toctree: ../stubs/

   SetLayout
   TrivialLayout
   DenseLayout
   SabreLayout
   CSPLayout
   VF2Layout
   ApplyLayout
   Layout2qDistance
   EnlargeWithAncilla
   FullAncillaAllocation
   SabrePreLayout

Routing
=======

.. autosummary::
   :toctree: ../stubs/

   BasicSwap
   LookaheadSwap
   StochasticSwap
   SabreSwap
   Commuting2qGateRouter
   StarPreRouting

Basis Change
============

.. autosummary::
   :toctree: ../stubs/

   BasisTranslator
   Decompose
   TranslateParameterizedGates
   Unroll3qOrMore
   UnrollCustomDefinitions

Optimizations
=============

.. autosummary::
   :toctree: ../stubs/

   Optimize1qGates
   Optimize1qGatesDecomposition
   Collect1qRuns
   Collect2qBlocks
   CollectMultiQBlocks
   CollectAndCollapse
   CollectLinearFunctions
   CollectCliffords
   ConsolidateBlocks
   CXCancellation
   InverseCancellation
   CommutationAnalysis
   CommutativeCancellation
   CommutativeInverseCancellation
   Optimize1qGatesSimpleCommutation
   RemoveDiagonalGatesBeforeMeasure
   RemoveResetInZeroState
   RemoveFinalReset
   HoareOptimizer
   TemplateOptimization
   EchoRZXWeylDecomposition
   ResetAfterMeasureSimplification
   OptimizeCliffords
   ElidePermutations
   NormalizeRXAngle
   OptimizeAnnotated
   Split2QUnitaries
   RemoveIdentityEquivalent

Calibration
=============

.. autosummary::
   :toctree: ../stubs/

   PulseGates
   RZXCalibrationBuilder
   RZXCalibrationBuilderNoEcho
   RXCalibrationBuilder

.. autofunction:: rzx_templates

Scheduling
=============

.. autosummary::
   :toctree: ../stubs/

   TimeUnitConversion
   ALAPScheduleAnalysis
   ASAPScheduleAnalysis
   PadDynamicalDecoupling
   PadDelay
   ConstrainedReschedule
   ValidatePulseGates
   InstructionDurationCheck
   SetIOLatency
   ALAPSchedule
   ASAPSchedule
   DynamicalDecoupling
   AlignMeasures

Circuit Analysis
================

.. autosummary::
   :toctree: ../stubs/

   Width
   Depth
   Size
   CountOps
   CountOpsLongestPath
   NumTensorFactors
   DAGLongestPath

Synthesis
=========

The synthesis transpiler plugin documentation can be found in the
:mod:`qiskit.transpiler.passes.synthesis.plugin` page.

.. autosummary::
   :toctree: ../stubs/

   UnitarySynthesis
   LinearFunctionsToPermutations
   HighLevelSynthesis
   HLSConfig
   SolovayKitaev

Post Layout
===========

These are post qubit selection.

.. autosummary::
   :toctree: ../stubs/

   VF2PostLayout

Additional Passes
=================

.. autosummary::
   :toctree: ../stubs/

   CheckMap
   CheckGateDirection
   GateDirection
   MergeAdjacentBarriers
   RemoveBarriers
   BarrierBeforeFinalMeasurements
   RemoveFinalMeasurements
   DAGFixedPoint
   FixedPoint
   MinimumPoint
   ContainsInstruction
   GatesInBasis
   ConvertConditionsToIfOps
   UnrollForLoops
   FilterOpNodes
"""

# layout selection (placement)
from .layout import SetLayout
from .layout import TrivialLayout
from .layout import DenseLayout
from .layout import SabreLayout
from .layout import CSPLayout
from .layout import VF2Layout
from .layout import VF2PostLayout
from .layout import ApplyLayout
from .layout import Layout2qDistance
from .layout import EnlargeWithAncilla
from .layout import FullAncillaAllocation
from .layout import SabrePreLayout

# routing
from .routing import BasicSwap
from .routing import LayoutTransformation
from .routing import LookaheadSwap
from .routing import StochasticSwap
from .routing import SabreSwap
from .routing import Commuting2qGateRouter
from .routing import StarPreRouting

# basis change
from .basis import Decompose
from .basis import UnrollCustomDefinitions
from .basis import Unroll3qOrMore
from .basis import BasisTranslator
from .basis import TranslateParameterizedGates

# optimization
from .optimization import Optimize1qGates
from .optimization import Optimize1qGatesDecomposition
from .optimization import Collect2qBlocks
from .optimization import Collect1qRuns
from .optimization import CollectMultiQBlocks
from .optimization import ConsolidateBlocks
from .optimization import CommutationAnalysis
from .optimization import CommutativeCancellation
from .optimization import CommutativeInverseCancellation
from .optimization import CXCancellation
from .optimization import Optimize1qGatesSimpleCommutation
from .optimization import OptimizeSwapBeforeMeasure
from .optimization import RemoveResetInZeroState
from .optimization import RemoveFinalReset
from .optimization import RemoveDiagonalGatesBeforeMeasure
from .optimization import HoareOptimizer
from .optimization import TemplateOptimization
from .optimization import InverseCancellation
from .optimization import EchoRZXWeylDecomposition
from .optimization import CollectAndCollapse
from .optimization import CollectLinearFunctions
from .optimization import CollectCliffords
from .optimization import ResetAfterMeasureSimplification
from .optimization import OptimizeCliffords
from .optimization import ElidePermutations
from .optimization import NormalizeRXAngle
from .optimization import OptimizeAnnotated
from .optimization import RemoveIdentityEquivalent
from .optimization import Split2QUnitaries

# circuit analysis
from .analysis import ResourceEstimation
from .analysis import Depth
from .analysis import Size
from .analysis import Width
from .analysis import CountOps
from .analysis import CountOpsLongestPath
from .analysis import NumTensorFactors
from .analysis import DAGLongestPath

# synthesis
from .synthesis import UnitarySynthesis
from .synthesis import unitary_synthesis_plugin_names
from .synthesis import LinearFunctionsToPermutations
from .synthesis import HighLevelSynthesis
from .synthesis import HLSConfig
from .synthesis import SolovayKitaev
from .synthesis import SolovayKitaevSynthesis
from .synthesis import AQCSynthesisPlugin

# calibration
from .calibration import PulseGates
from .calibration import RZXCalibrationBuilder
from .calibration import RZXCalibrationBuilderNoEcho
from .calibration import RXCalibrationBuilder
from .calibration.rzx_templates import rzx_templates

# circuit scheduling
from .scheduling import TimeUnitConversion
from .scheduling import ALAPScheduleAnalysis
from .scheduling import ASAPScheduleAnalysis
from .scheduling import PadDynamicalDecoupling
from .scheduling import ValidatePulseGates
from .scheduling import PadDelay
from .scheduling import ConstrainedReschedule
from .scheduling import InstructionDurationCheck
from .scheduling import SetIOLatency
from .scheduling import ALAPSchedule
from .scheduling import ASAPSchedule
from .scheduling import DynamicalDecoupling
from .scheduling import AlignMeasures

# additional utility passes
from .utils import CheckMap
from .utils import CheckGateDirection
from .utils import GateDirection
from .utils import BarrierBeforeFinalMeasurements
from .utils import RemoveFinalMeasurements
from .utils import MergeAdjacentBarriers
from .utils import DAGFixedPoint
from .utils import FixedPoint
from .utils import MinimumPoint
from .utils import Error
from .utils import RemoveBarriers
from .utils import ContainsInstruction
from .utils import GatesInBasis
from .utils import ConvertConditionsToIfOps
from .utils import UnrollForLoops
from .utils import FilterOpNodes



# This code is part of Qiskit.
#
# (C) Copyright IBM 2017, 2024.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Tests for Clifford synthesis functions."""

from ddt import ddt
from qiskit.quantum_info import random_clifford
from qiskit.quantum_info.operators import Clifford
from qiskit.synthesis.clifford import (
    synth_clifford_full,
    synth_clifford_ag,
    synth_clifford_bm,
    synth_clifford_greedy,
)

from test import QiskitTestCase  # pylint: disable=wrong-import-order
from test import combine  # pylint: disable=wrong-import-order


@ddt
class TestCliffordSynthesis(QiskitTestCase):
    """Tests for clifford synthesis functions."""

    @staticmethod
    def _cliffords_1q():
        clifford_dicts = [
            {"stabilizer": ["+Z"], "destabilizer": ["-X"]},
            {"stabilizer": ["-Z"], "destabilizer": ["+X"]},
            {"stabilizer": ["-Z"], "destabilizer": ["-X"]},
            {"stabilizer": ["+Z"], "destabilizer": ["+Y"]},
            {"stabilizer": ["+Z"], "destabilizer": ["-Y"]},
            {"stabilizer": ["-Z"], "destabilizer": ["+Y"]},
            {"stabilizer": ["-Z"], "destabilizer": ["-Y"]},
            {"stabilizer": ["+X"], "destabilizer": ["+Z"]},
            {"stabilizer": ["+X"], "destabilizer": ["-Z"]},
            {"stabilizer": ["-X"], "destabilizer": ["+Z"]},
            {"stabilizer": ["-X"], "destabilizer": ["-Z"]},
            {"stabilizer": ["+X"], "destabilizer": ["+Y"]},
            {"stabilizer": ["+X"], "destabilizer": ["-Y"]},
            {"stabilizer": ["-X"], "destabilizer": ["+Y"]},
            {"stabilizer": ["-X"], "destabilizer": ["-Y"]},
            {"stabilizer": ["+Y"], "destabilizer": ["+X"]},
            {"stabilizer": ["+Y"], "destabilizer": ["-X"]},
            {"stabilizer": ["-Y"], "destabilizer": ["+X"]},
            {"stabilizer": ["-Y"], "destabilizer": ["-X"]},
            {"stabilizer": ["+Y"], "destabilizer": ["+Z"]},
            {"stabilizer": ["+Y"], "destabilizer": ["-Z"]},
            {"stabilizer": ["-Y"], "destabilizer": ["+Z"]},
            {"stabilizer": ["-Y"], "destabilizer": ["-Z"]},
        ]
        return [Clifford.from_dict(i) for i in clifford_dicts]

    def test_decompose_1q(self):
        """Test synthesis for all 1-qubit Cliffords"""
        for cliff in self._cliffords_1q():
            with self.subTest(msg=f"Test circuit {cliff}"):
                target = cliff
                value = Clifford(cliff.to_circuit())
                self.assertEqual(target, value)

    @combine(num_qubits=[1, 2, 3])
    def test_synth_bm(self, num_qubits):
        """Test B&M synthesis for set of {num_qubits}-qubit Cliffords"""
        samples = 50
        for seed in range(samples):
            target = random_clifford(num_qubits, seed=seed)
            synth_circ = synth_clifford_bm(target)
            value = Clifford(synth_circ)
            self.assertEqual(value, target)

    @combine(num_qubits=[1, 2, 3, 4, 5])
    def test_synth_ag(self, num_qubits):
        """Test A&G synthesis for set of {num_qubits}-qubit Cliffords"""
        samples = 50
        for seed in range(samples):
            target = random_clifford(num_qubits, seed)
            synth_circ = synth_clifford_ag(target)
            value = Clifford(synth_circ)
            self.assertEqual(value, target)

    @combine(num_qubits=[1, 2, 3, 4, 5])
    def test_synth_greedy(self, num_qubits):
        """Test greedy synthesis for set of {num_qubits}-qubit Cliffords"""
        samples = 50
        for seed in range(samples):
            target = random_clifford(num_qubits, seed)
            synth_circ = synth_clifford_greedy(target)
            value = Clifford(synth_circ)
            self.assertEqual(value, target)

    @combine(num_qubits=[1, 2, 3, 4, 5])
    def test_synth_full(self, num_qubits):
        """Test synthesis for set of {num_qubits}-qubit Cliffords"""
        samples = 50
        for seed in range(samples):
            target = random_clifford(num_qubits, seed)
            synth_circ = synth_clifford_full(target)
            value = Clifford(synth_circ)
            self.assertEqual(value, target)



# This code is part of Qiskit.
#
# (C) Copyright IBM 2019.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.


"""
Global Mølmer–Sørensen gate.
"""

from __future__ import annotations
from collections.abc import Sequence

import numpy as np
from qiskit.circuit.quantumcircuit import QuantumCircuit
from qiskit.circuit.quantumregister import QuantumRegister
from qiskit.circuit.parameterexpression import ParameterValueType
from qiskit.circuit.library.standard_gates import RXXGate
from qiskit.circuit.gate import Gate
from qiskit.utils.deprecation import deprecate_func


class GMS(QuantumCircuit):
    r"""Global Mølmer–Sørensen gate.

    **Circuit symbol:**

    .. code-block:: text

             ┌───────────┐
        q_0: ┤0          ├
             │           │
        q_1: ┤1   GMS    ├
             │           │
        q_2: ┤2          ├
             └───────────┘

    **Expanded Circuit:**

    .. plot::
       :alt: Diagram illustrating the previously described circuit.

       from qiskit.circuit.library import GMS
       from qiskit.visualization.library import _generate_circuit_library_visualization
       import numpy as np
       circuit = GMS(num_qubits=3, theta=[[0, np.pi/4, np.pi/8],
                                          [0, 0, np.pi/2],
                                          [0, 0, 0]])
       _generate_circuit_library_visualization(circuit.decompose())

    The Mølmer–Sørensen gate is native to ion-trap systems. The global MS
    can be applied to multiple ions to entangle multiple qubits simultaneously [1].

    In the two-qubit case, this is equivalent to an XX(theta) interaction,
    and is thus reduced to the RXXGate. The global MS gate is a sum of XX
    interactions on all pairs [2].

    .. math::

        GMS(\chi_{12}, \chi_{13}, ..., \chi_{n-1 n}) =
        exp(-i \sum_{i=1}^{n} \sum_{j=i+1}^{n} X{\otimes}X \frac{\chi_{ij}}{2})

    **References:**

    [1] Sørensen, A. and Mølmer, K., Multi-particle entanglement of hot trapped ions.
    Physical Review Letters. 82 (9): 1835–1838.
    `arXiv:9810040 <https://arxiv.org/abs/quant-ph/9810040>`_

    [2] Maslov, D. and Nam, Y., Use of global interactions in efficient quantum circuit
    constructions. New Journal of Physics, 20(3), p.033018.
    `arXiv:1707.06356 <https://arxiv.org/abs/1707.06356>`_
    """

    @deprecate_func(since="1.3", additional_msg="Use the MSGate instead.", pending=True)
    def __init__(self, num_qubits: int, theta: list[list[float]] | np.ndarray) -> None:
        """Create a new Global Mølmer–Sørensen (GMS) gate.

        Args:
            num_qubits: width of gate.
            theta: a num_qubits x num_qubits symmetric matrix of
                interaction angles for each qubit pair. The upper
                triangle is considered.
        """
        super().__init__(num_qubits, name="gms")
        if not isinstance(theta, list):
            theta = [theta] * int((num_qubits**2 - 1) / 2)
        gms = QuantumCircuit(num_qubits, name="gms")
        for i in range(self.num_qubits):
            for j in range(i + 1, self.num_qubits):
                gms.append(RXXGate(theta[i][j]), [i, j])
        self.append(gms.to_gate(), self.qubits)


class MSGate(Gate):
    r"""The Mølmer–Sørensen gate.

    The Mølmer–Sørensen gate is native to ion-trap systems. The global MS
    can be applied to multiple ions to entangle multiple qubits simultaneously [1].

    In the two-qubit case, this is equivalent to an XX interaction,
    and is thus reduced to the :class:`.RXXGate`. The global MS gate is a sum of XX
    interactions on all pairs [2].

    .. math::

        MS(\chi_{12}, \chi_{13}, ..., \chi_{n-1 n}) =
        exp(-i \sum_{i=1}^{n} \sum_{j=i+1}^{n} X{\otimes}X \frac{\chi_{ij}}{2})

    Example::

        import numpy as np
        from qiskit.circuit.library import MSGate
        from qiskit.quantum_info import Operator

        gate = MSGate(num_qubits=3, theta=[[0, np.pi/4, np.pi/8],
                                           [0, 0, np.pi/2],
                                           [0, 0, 0]])
        print(Operator(gate))


    **References:**

    [1] Sørensen, A. and Mølmer, K., Multi-particle entanglement of hot trapped ions.
    Physical Review Letters. 82 (9): 1835–1838.
    `arXiv:9810040 <https://arxiv.org/abs/quant-ph/9810040>`_

    [2] Maslov, D. and Nam, Y., Use of global interactions in efficient quantum circuit
    constructions. New Journal of Physics, 20(3), p.033018.
    `arXiv:1707.06356 <https://arxiv.org/abs/1707.06356>`_
    """

    def __init__(
        self,
        num_qubits: int,
        theta: ParameterValueType | Sequence[Sequence[ParameterValueType]],
        label: str | None = None,
    ):
        """
        Args:
            num_qubits: The number of qubits the MS gate acts on.
            theta: The XX rotation angles. If a single value, the same angle is used on all
                interactions. Alternatively an upper-triangular, square matrix with width
                ``num_qubits`` can be provided with interaction angles for each qubit pair.
            label: A gate label.
        """
        super().__init__("ms", num_qubits, [theta], label=label)

    def _define(self):
        thetas = self.params[0]
        q = QuantumRegister(self.num_qubits, name="q")
        qc = QuantumCircuit(q, name=self.name)
        for i in range(self.num_qubits):
            for j in range(i + 1, self.num_qubits):
                # if theta is just a single angle, use that, otherwise use the correct index
                theta = thetas if not isinstance(thetas, Sequence) else thetas[i][j]
                qc._append(RXXGate(theta), [q[i], q[j]], [])

        self.definition = qc

    def validate_parameter(self, parameter):
        if isinstance(parameter, Sequence):
            # pylint: disable=super-with-arguments
            return [
                [super(MSGate, self).validate_parameter(theta) for theta in row]
                for row in parameter
            ]

        return super().validate_parameter(parameter)
